#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
API endpoint that deliver backend configuration details.
'''

from typing import Dict, Any, List, Optional
from datetime import datetime
from fastapi.routing import APIRouter
from pydantic.fields import Field
from pydantic.main import BaseModel

from nomad import config, normalizing, gitinfo
from nomad.utils import strip
from nomad.search import search
from nomad.parsing import parsers, MatchingParser
from nomad.app.v1.models import Aggregation, StatisticsAggregation
from nomad.metainfo.elasticsearch_extension import entry_type


router = APIRouter()
default_tag = 'info'


class MetainfoModel(BaseModel):
    all_package: str = Field(None, description=strip('''
        Name of the metainfo package that references all available packages, i.e.
        the complete metainfo.'''))

    root_section: str = Field(None, description=strip('''
        Name of the topmost section, e.g. section run for computational material science
        data.'''))


class GitInfoModel(BaseModel):
    ref: str
    version: str
    commit: str
    log: str


class StatisticsModel(BaseModel):
    n_entries: int = Field(None, description='Number of entries in NOMAD')
    n_uploads: int = Field(None, description='Number of uploads in NOMAD')
    n_quantities: int = Field(None, description='Accumulated number of quantities over all entries in the Archive')
    n_calculations: int = Field(None, description='Accumulated number of calculations, e.g. total energy calculations in the Archive')
    n_materials: int = Field(None, description='Number of materials in NOMAD')
    # TODO raw_file_size, archive_file_size


class CodeInfoModel(BaseModel):
    code_name: Optional[str] = Field(None, description='Name of the code or input format')
    code_homepage: Optional[str] = Field(None, description='Homepage of the code or input format')


class InfoModel(BaseModel):
    parsers: List[str]
    metainfo_packages: List[str]
    codes: List[CodeInfoModel]
    normalizers: List[str]
    statistics: StatisticsModel = Field(None, description='General NOMAD statistics')
    search_quantities: dict
    version: str
    deployment: str
    git: GitInfoModel
    oasis: bool


_statistics: Dict[str, Any] = None


def statistics():
    global _statistics
    if _statistics is None or datetime.now().timestamp() - _statistics.get('timestamp', 0) > 3600 * 24:
        _statistics = dict(timestamp=datetime.now().timestamp())
        search_response = search(aggregations=dict(statistics=Aggregation(statistics=StatisticsAggregation(
            metrics=['n_entries', 'n_materials', 'n_uploads', 'n_quantities', 'n_calculations']))))
        _statistics.update(**search_response.aggregations['statistics'].statistics.data)  # pylint: disable=no-member

    return _statistics


@router.get(
    '',
    tags=[default_tag],
    summary='Get information about the nomad backend and its configuration',
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    response_model=InfoModel)
async def get_info():
    ''' Return information about the nomad backend and its configuration. '''
    codes_dict = {}
    for parser in parsers.parser_dict.values():
        if isinstance(parser, MatchingParser) and parser.domain == 'dft':
            code_name = parser.code_name
            if code_name in codes_dict:
                continue
            codes_dict[code_name] = dict(code_name=code_name, code_homepage=parser.code_homepage)
    codes = sorted(list(codes_dict.values()), key=lambda code_info: code_info['code_name'].lower())

    return {
        'parsers': [
            key[key.index('/') + 1:]
            for key in parsers.parser_dict.keys()],
        'metainfo_packages': ['general', 'general.experimental', 'common', 'public'] + sorted([
            key[key.index('/') + 1:]
            for key in parsers.parser_dict.keys()]),
        'codes': codes,
        'normalizers': [normalizer.__name__ for normalizer in normalizing.normalizers],
        'statistics': statistics(),
        'search_quantities': {
            s.qualified_name: {
                'name': s.qualified_name,
                'description': s.definition.description,
                'many': not s.definition.is_scalar
            }
            for s in entry_type.quantities.values()
            if 'optimade' not in s.qualified_name
        },
        'version': config.meta.version,
        'deployment': config.meta.deployment,
        'git': {
            'ref': gitinfo.ref,
            'version': gitinfo.version,
            'commit': gitinfo.commit,
            'log': gitinfo.log
        },
        'oasis': config.keycloak.oasis
    }




#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click
import sys

from nomad import config
from nomad.cli.cli import cli


@cli.group(help='''The nomad admin commands to do nasty stuff directly on the databases.
                     Remember: With great power comes great responsibility!''')
@click.pass_context
def admin(ctx):
    pass


@admin.command(help='Reset/remove all databases.')
@click.option('--remove', is_flag=True, help='Do not just reset all dbs, but also remove them.')
@click.option('--i-am-really-sure', is_flag=True, help='Must be set for the command to to anything.')
def reset(remove, i_am_really_sure):
    if not i_am_really_sure:
        print('You do not seem to be really sure about what you are doing.')
        sys.exit(1)

    from nomad import infrastructure

    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    infrastructure.reset(remove)


@admin.command(help='Reset all uploads and entries "stuck" in processing using level mongodb operations.')
@click.option('--zero-complete-time', is_flag=True, help='Sets the complete time to epoch zero.')
def reset_processing(zero_complete_time):
    from datetime import datetime

    from nomad import infrastructure, processing as proc

    infrastructure.setup_mongo()

    def reset_collection(cls):
        in_processing = cls.objects(process_status__in=proc.ProcessStatus.STATUSES_PROCESSING)
        print('%d %s processes need to be reset due to incomplete process' % (in_processing.count(), cls.__name__))
        in_processing.update(
            process_status=proc.ProcessStatus.READY,
            current_process=None,
            worker_hostname=None,
            celery_task_id=None,
            errors=[], warnings=[],
            complete_time=datetime.fromtimestamp(0) if zero_complete_time else datetime.now())

    reset_collection(proc.Entry)
    reset_collection(proc.Upload)


@admin.command(help='Check and lift embargo of data with expired embargo period.')
@click.option('--dry', is_flag=True, help='Do not lift the embargo, just show what needs to be done.')
@click.option('--parallel', default=1, type=int, help='Use the given amount of parallel processes. Default is 1.')
def lift_embargo(dry, parallel):
    from datetime import datetime
    from dateutil.relativedelta import relativedelta

    from nomad import infrastructure, processing as proc
    from nomad.search import quantity_values

    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    query = dict(with_embargo=True, published=True)

    for upload_id in quantity_values('upload_id', query=query, owner='all'):
        upload = proc.Upload.get(upload_id)
        embargo_length = upload.embargo_length

        if upload.publish_time + relativedelta(months=embargo_length) < datetime.now():
            print('need to lift the embargo of %s (publish_time=%s, embargo=%d)' % (
                upload.upload_id, upload.publish_time, embargo_length))

            if not dry:
                upload.edit_upload_metadata(
                    edit_request_json=dict(metadata={'embargo_length': 0}),
                    user_id=config.services.admin_user_id)
    return


@admin.group(help='Generate scripts and commands for nomad operation.')
def ops():
    pass


# @ops.group(help='Tools for managing the DOS similarity data.')
# def similarity():
#     pass


@ops.command(help=('Dump the mongo db.'))
@click.option('--restore', is_flag=True, help='Do not dump, but restore.')
def dump(restore: bool):
    from datetime import datetime

    date_str = datetime.utcnow().strftime('%Y_%m_%d')
    print('mongodump --host {} --port {} --db {} -o /backup/fairdi/mongo/{}'.format(
        config.mongo.host, config.mongo.port, config.mongo.db_name, date_str))


@ops.command(help=('Restore the mongo db.'))
@click.argument('PATH_TO_DUMP', type=str, nargs=1)
def restore(path_to_dump):
    print('mongorestore --host {} --port {} --db {} {}'.format(
        config.mongo.host, config.mongo.port, config.mongo.db_name, path_to_dump))


@ops.command(help=('Generate an nginx.conf to serve the GUI and proxy pass to API container.'))
@click.option('--prefix', type=str, default=config.services.api_base_path, help='Alter the url path prefix.')
@click.option('--host', type=str, default=config.services.api_host, help='Alter the NOMAD app host.')
@click.option('--port', type=str, default=config.services.api_port, help='Alter the NOMAD port host.')
@click.option('--server/--no-server', default=True, help='Control writing of the outer server {} block. '
              'Useful when conf file is included within another nginx.conf.')
def nginx_conf(prefix, host, port, server):
    prefix = prefix.rstrip('/')
    prefix = '/%s' % prefix.lstrip('/')

    if server:
        print('''server {
    listen        80;
    server_name   www.example.com;
    proxy_set_header Host $host;
        ''')

    print('''
    location / {{
        proxy_pass http://{1}:{2};
    }}

    location ~ {0}\\/?(gui)?$ {{
        rewrite ^ {0}/gui/ permanent;
    }}

    location {0}/gui/ {{
        proxy_intercept_errors on;
        error_page 404 = @redirect_to_index;
        proxy_pass http://{1}:{2};
    }}

    location @redirect_to_index {{
        rewrite ^ {0}/gui/index.html break;
        proxy_pass http://{1}:{2};
    }}

    location ~ \\/gui\\/(service-worker\\.js|meta\\.json)$ {{
        add_header Last-Modified $date_gmt;
        add_header Cache-Control 'no-store, no-cache, must-revalidate, proxy-revalidate, max-age=0';
        if_modified_since off;
        expires off;
        etag off;
        proxy_pass http://{1}:{2};
    }}

    location ~ /api/v1/uploads(/?$|.*/raw|.*/bundle?$) {{
        client_max_body_size 35g;
        proxy_request_buffering off;
        proxy_pass http://{1}:{2};
    }}

    location ~ /api/v1/.*/download {{
        proxy_buffering off;
        proxy_pass http://{1}:{2};
    }}
'''.format(prefix, host, port))
    if server:
        print('}')


@ops.command(help='Updates the AFLOW prototype information using the latest online version and writes the results to a python module in the given FILEPATH.')
@click.argument('FILEPATH', nargs=1, type=str)
@click.option('--matches-only', is_flag=True, help='Only update the match information that depends on the symmetry analysis settings. Will not perform and online update.')
@click.pass_context
def prototypes_update(ctx, filepath, matches_only):
    from nomad.cli.aflow import update_prototypes
    update_prototypes(ctx, filepath, matches_only)


@ops.command(help='Updates the springer database in nomad.config.normalize.springer_db_path.')
@click.option('--max-n-query', default=10, type=int, help='Number of unsuccessful springer request before returning an error. Default is 10.')
@click.option('--retry-time', default=120, type=int, help='Time in seconds to retry after unsuccessful request. Default is 120.')
def springer_update(max_n_query, retry_time):
    from nomad.cli.admin import springer
    springer.update_springer(max_n_query, retry_time)


# @similarity.command(help='Updates the msgpack file containing the similarity information.')
# @click.option('--dir', "-d", "input_dir", type=str, help='Path of the folder containing the raw similarity information files')
# @click.option('--out', "-o", type=str, help='Path of the output msgpack file.')
# @click.option('--verbose', is_flag=True, help='Enable verbose output.')
# def update(input_dir, out, verbose):
#     from nomad.cli.admin import similarity
#     similarity.update(input_dir, out, verbose)


# @similarity.command(help='Ingests the given similarity information from an msgpack file into MongoDB.')
# @click.option('--in', "-i", "input_path", type=str, help='Path of the ingested msgpack file.')
# @click.option('--batch_size', type=int, default=10000, help='Batch size for MongoDB bulk ingestion.')
# @click.option('--verbose', is_flag=True, help='Enable verbose output.')
# def ingest(input_path, batch_size, verbose):
#     from nomad.cli.admin import similarity
#     similarity.ingest(input_path, batch_size, verbose)


@ops.command(help='Configures the GUIs based on NOMAD config.')
def gui_config():
    import os
    import os.path
    from nomad import config
    import glob
    import shutil
    import json

    gui_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../app/flask/static/gui'))
    run_gui_folder = os.path.join(gui_folder, '../.gui_configured')

    # copy
    shutil.rmtree(run_gui_folder, ignore_errors=True)
    shutil.copytree(gui_folder, run_gui_folder)

    # setup the env
    env_js_file = os.path.join(run_gui_folder, 'env.js')
    if not os.path.exists(env_js_file):
        with open(env_js_file, 'wt') as f:
            f.write(f'''
window.nomadEnv = {{
    'appBase': '{config.services.api_base_path.rstrip("/")}',
    'northBase': '{config.services.api_base_path.rstrip("/")}/north',
    'keycloakBase': '{config.keycloak.public_server_url}',
    'keycloakRealm': '{config.keycloak.realm_name}',
    'keycloakClientId': '{config.keycloak.client_id}',
    'debug': false,
    'encyclopediaBase': '{config.encyclopedia_base if config.encyclopedia_base else 'undefined'}',
    'aitoolkitEnabled': {'true' if config.aitoolkit_enabled else 'false'},
    'oasis': {'true' if config.keycloak.oasis else 'false'},
    'version': {json.dumps(config.meta.beta) if config.meta.beta else dict()},
    'globalLoginRequired': {'false' if config.oasis.allowed_users is None else 'true'}
}};''')

    # replace base path in all GUI files
    source_file_globs = [
        '**/*.json',
        '**/*.html',
        '**/*.js',
        '**/*.js.map',
        '**/*.css']
    for source_file_glob in source_file_globs:
        source_files = glob.glob(os.path.join(run_gui_folder, source_file_glob), recursive=True)
        for source_file in source_files:
            with open(source_file, 'rt') as f:
                file_data = f.read()
            file_data = file_data.replace('/fairdi/nomad/latest', config.services.api_base_path)
            with open(source_file, 'wt') as f:
                f.write(file_data)


@admin.group(help='Commands for upgrading to a newer NOMAD version')
def upgrade():
    pass


@upgrade.command(
    help='''Converts (upgrades) records from one mongodb and migrates to another.
            Note, it is strongly recommended to run this command with loglevel verbosed, i.e.

                nomad -v upgrade migrate-mongo ...

         ''')
@click.option(
    '--host', type=str, default=config.mongo.host,
    help='The mongodb host. By default same as the configureed db.')
@click.option(
    '--port', type=int, default=config.mongo.port,
    help='The mongodb port. By default same as the configured db.')
@click.option(
    '--src-db-name', type=str, required=True,
    help='The name of the source database.')
@click.option(
    '--dst-db-name', type=str, default=config.mongo.db_name,
    help='The name of the destination database. By default same as the configured db.')
@click.option(
    '--upload-query', type=str,
    help='An mongo upload query. All uploads matching the query will be included in the migration.')
@click.option(
    '--entry-query', type=str,
    help='An mongo entry query. All uploads with an entry matching the query will be included in the migration.')
@click.option(
    '--ids-from-file', type=str,
    help='''Reads upload IDs from the specified file. Cannot be used together with the
            --upload-query or --entry-query options.
            This can for example be used to retry just the uploads that has previously failed
            (as these ids can be exported to file using --failed-ids-to-file). You can specify both
            --ids-from-file and --failed-ids-to-file at the same time with the same file name.''')
@click.option(
    '--failed-ids-to-file', type=str,
    help='''Write the IDs of failed and skipped uploads to the specified file.
            This can for example be used to subsequently retry just the uploads that failed
            (as these ids can be loaded from file using --ids-from-file). You can specify both
            --ids-from-file and --failed-ids-to-file at the same time with the same file name.''')
@click.option(
    '--upload-update', type=str,
    help='json with updates to apply to all converted uploads')
@click.option(
    '--entry-update', type=str,
    help='json with updates to apply to all converted entries')
@click.option(
    '--overwrite', type=click.Choice(['always', 'if-newer', 'never'], case_sensitive=False), default='never',
    help='''If an upload already exists in the destination db, this option determines whether
            it and its child records should be overwritten with the data from the source db.
            Possible values are "always", "if-newer", "never". Selecting "always" always overwrites,
            "never" never overwrites, and "if-newer" overwrites if the upload either doesn't exist
            in the destination, or it exists but its complete_time (i.e. last time it was
            processed) is older than in the source db.''')
@click.option(
    '--fix-problems', is_flag=True,
    help='''If a minor, fixable problem is encountered, fixes it automaticall; otherwise fail.''')
@click.option(
    '--dry', is_flag=True,
    help='Dry run (not writing anything to the destination database).')
def migrate_mongo(
        host, port, src_db_name, dst_db_name, upload_query, entry_query,
        ids_from_file, failed_ids_to_file, upload_update, entry_update, overwrite, fix_problems, dry):
    import json
    from pymongo.database import Database
    from nomad import infrastructure
    import nomad.cli.admin.migrate as migrate

    config.mongo.host = host
    config.mongo.port = port
    config.mongo.db_name = dst_db_name
    infrastructure.setup_mongo()

    db_src: Database = infrastructure.mongo_client.get_database(src_db_name)
    db_dst: Database = infrastructure.mongo_client.get_database(dst_db_name)

    if not dry:
        migrate.create_collections_if_needed(db_dst)

    upload_ids = None
    if upload_query and entry_query:
        print('Cannot specify both upload-query and entry-query')
        return -1
    if ids_from_file:
        if upload_query or entry_query:
            print('Cannot specify a query when using --ids-from-file.')
            return -1
        try:
            with open(ids_from_file, 'r') as f:
                upload_ids = [line.strip() for line in f.readlines() if line.strip()]
        except FileNotFoundError:
            print(f'Could not open file {ids_from_file}', file=sys.stderr)
            return -1
    elif upload_query:
        upload_query = json.loads(upload_query)
    elif entry_query:
        entry_query = json.loads(entry_query)

    if upload_update:
        upload_update = json.loads(upload_update)
    if entry_update:
        entry_update = json.loads(entry_update)

    if entry_query:
        print('Quering entries...')
        src_entry_collection = db_src.calc if 'calc' in db_src.collection_names() else db_src.entry
        upload_ids = list(src_entry_collection.distinct('upload_id', entry_query))
    if upload_ids:
        upload_query = {'_id': {'$in': upload_ids}}
    print('Quering uploads...')
    uploads = db_src.upload.find(upload_query)

    migrate.migrate_mongo_uploads(
        db_src, db_dst, uploads, failed_ids_to_file, upload_update, entry_update, overwrite,
        fix_problems, dry)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os.path
import os
import json
from datetime import datetime
import logging
import glob
import re
import numpy as np

from nomad.datamodel import EntryArchive
from nomad.datamodel.results import Material
from nomad.metainfo.metainfo import SubSection
from nomad.parsing.parser import MatchingParser
from nomad.units import ureg
from nomad.metainfo import MSection, Section, Quantity, Datetime, Package
from nomad.metainfo.elasticsearch_extension import Elasticsearch, material_entry_type

from nomad.datamodel.metainfo.measurements import (
    Measurement, Sample, Instrument, Spectrum)
from nomad.datamodel import Author
from nomad.datamodel.results import (
    Results, Method, Properties, SpectroscopyProperties)


default_logger = logging.getLogger(__name__)

m_package = Package(name='eels')


class EELSMeasurement(MSection):
    # m_def = Section(parents=[
    #     ParentSection(name='eels', section_def=Measurement)])

    spectrum = SubSection(section_def=Spectrum)
    edges = Quantity(type=str, shape=['*'])
    publish_time = Quantity(
        type=Datetime, description='The datetime that this was published on EELS DB.')
    authors = SubSection(section_def=Author, repreats=True)


class EELSInstrument(MSection):
    # m_def = Section(parents=[
    #     ParentSection(name='eels', section_def=Instrument)])

    max_energy = Quantity(type=np.dtype(np.float64), unit="joule")
    min_energy = Quantity(type=np.dtype(np.float64), unit="joule")
    guntype = Quantity(type=str)
    beam_energy = Quantity(type=str)
    resolution = Quantity(type=np.dtype(np.float64), unit="joule")
    step_size = Quantity(type=str)
    acquisition_mode = Quantity(type=str)
    beam_current = Quantity(type=str)
    detector_type = Quantity(type=str)
    dark_current = Quantity(type=str)


class EELSProperty(MSection):
    detector_type = EELSInstrument.detector_type.m_copy(es_annotation=[
        Elasticsearch(material_entry_type),
        Elasticsearch(suggestion="default")
    ])
    resolution = EELSInstrument.resolution.m_copy(es_annotation=[
        Elasticsearch(material_entry_type),
    ])
    max_energy = EELSInstrument.max_energy.m_copy(es_annotation=[
        Elasticsearch(material_entry_type),
    ])
    min_energy = EELSInstrument.min_energy.m_copy(es_annotation=[
        Elasticsearch(material_entry_type),
    ])


# TODO replace with ParentSection once this is implement in the metainfo
class MyMeasurement(Measurement):
    m_def = Section(extends_base_section=True)
    eels = SubSection(section_def=EELSMeasurement)


class MyInstrument(Instrument):
    m_def = Section(extends_base_section=True)
    eels = SubSection(section_def=EELSInstrument)


class MySpectroscopyProperties(SpectroscopyProperties):
    m_def = Section(extends_base_section=True)
    eels = SubSection(
        sub_section=EELSProperty,
        a_elasticsearch=Elasticsearch(material_entry_type, nested=True))


m_package.__init_metainfo__()


class EELSDBParser(MatchingParser):
    def __init__(self):
        super().__init__(
            name='parsers/eels', code_name='eels', code_homepage='https://eelsdb.eu/',
            domain='ems',
            mainfile_mime_re=r'application/json',
            mainfile_contents_re=(r'https://eelsdb.eu/spectra')
        )

        self.logger = None

    def parse_msa_file(self, msa_path) -> Spectrum:
        ''' Parses the given msa file and returns a spectrum with data from that file. '''

        metadata_re = re.compile(r'^#\s*([A-Z0-9]+)\s*:(.*)\s*$')
        number_re = r'-?\d+(\.\d+)?(e[\+\-]?\d+)?'
        data_re = re.compile(f'({number_re}),\\s*({number_re})')

        metadata = {}
        energies = []
        counts = []
        with open(msa_path, 'rt') as f:
            for line in f.readlines():
                match = re.match(metadata_re, line)
                if match:
                    metadata[match.group(1)] = match.group(2)
                    continue
                match = re.match(data_re, line)
                if match:
                    energies.append(float(match.group(1)))
                    counts.append(float(match.group(4)))
                    continue

                self.logger.warning('Unexpected line format in .msa file')

        x_units = metadata.get('XUNITS')
        if not x_units or 'undefined' in x_units.lower():
            x_units = 'eV'
            self.logger.warning('Unknown energy units')

        if int(metadata.get('NPOINTS', 0)) != len(energies):
            self.logger.warning('Npoints metadata does not match value count')

        spectrum = Spectrum()
        spectrum.energy = energies * ureg(x_units)
        spectrum.count = counts
        return spectrum

    def parse(self, mainfile_path, archive: EntryArchive, logger=None):
        self.logger = logger
        if not self.logger:
            self.logger = default_logger

        with open(mainfile_path, 'rt') as f:
            raw_metadata = json.load(f)

        measurement = archive.m_create(Measurement)
        measurement.eels = EELSMeasurement()

        # Data
        msa_path = next(iter(
            glob.glob(os.path.join(os.path.dirname(mainfile_path), '*.msa'))), None)
        if msa_path:
            measurement.eels.spectrum = self.parse_msa_file(msa_path)
        else:
            logger.warning('No *.msa file found')

        # Sample
        sample = Sample()
        measurement.sample.append(sample)
        sample.chemical_formula = raw_metadata['formula']
        sample.name = raw_metadata['title']
        elements = raw_metadata.get('elements', [])
        if isinstance(elements, str):
            elements = json.loads(elements)
        sample.elements = elements

        # Measurement
        measurement.measurement_id = str(raw_metadata['id'])
        archive.metadata.external_id = str(raw_metadata['id'])
        measurement.method_name = 'electron energy loss spectroscopy'
        measurement.method_abbreviation = 'EELS'
        measurement.eels.publish_time = datetime.strptime(
            raw_metadata.get('published'), '%Y-%m-%d %H:%M:%S')
        edges = raw_metadata.get('edges', [])
        if isinstance(edges, str):
            edges = json.loads(edges)
        measurement.eels.edges = edges
        measurement.description = raw_metadata['description']

        # Instrument
        # TODO: Add units to variables here
        instrument = measurement.m_create(Instrument)
        instrument.eels = EELSInstrument()
        instrument.name = raw_metadata['microscope']
        max_energy_string = raw_metadata.get('max_energy')
        if max_energy_string is not None:
            value, unit = max_energy_string.split()
            instrument.eels.max_energy = float(value) * ureg(unit)
        min_energy_string = raw_metadata.get('min_energy')
        if min_energy_string is not None:
            value, unit = min_energy_string.split()
            instrument.eels.min_energy = float(value) * ureg(unit)
        resolution_string = raw_metadata.get('resolution')
        if resolution_string is not None:
            value, unit = resolution_string.split()
            instrument.eels.resolution = float(value) * ureg(unit)
        instrument.eels.guntype = raw_metadata['guntype']
        instrument.eels.beam_energy = raw_metadata['beamenergy']
        instrument.eels.step_size = raw_metadata['stepSize']
        if raw_metadata.get('acquisition_mode') is not None:
            instrument.eels.acquisition_mode = raw_metadata['acquisition_mode']
        if raw_metadata.get('beamcurrent') is not None:
            instrument.eels.beam_current = raw_metadata['beamcurrent']
        instrument.eels.detector_type = raw_metadata['detector']
        instrument.eels.dark_current = raw_metadata['darkcurrent']

        # Origin
        archive.metadata.external_db = 'EELS Data Base'
        archive.metadata.references = [
            raw_metadata[reference]
            for reference in ['permalink', 'preview_url', 'entry_repository_url', 'api_permalink']
            if reference in raw_metadata]

        # # Author
        author = measurement.eels.m_create(Author)
        names = raw_metadata['author']['name'].rsplit(' ', 1)
        if len(names) == 2:
            author.first_name, author.last_name = names
        else:
            author.last_name = names[0]
        archive.metadata.entry_coauthors = [author]
        # author.author_profile_url = raw_metadata['author']['profile_url']
        # author.author_profile_api_url = raw_metadata['author']['profile_api_url']

        # Results
        if archive.results is None:
            archive.results = Results()
        results = archive.results

        if results.material is None:
            results.material = Material()
        results.material.elements = sample.elements
        results.material.chemical_formula_descriptive = sample.chemical_formula

        if results.method is None:
            results.m_create(Method)
        results.method.method_name = 'EELS'

        if measurement.eels.spectrum:
            if results.properties is None:
                properties = results.m_create(Properties)
            if results.properties.spectroscopy is None:
                spectroscopy = properties.m_create(SpectroscopyProperties)

            spectroscopy.spectrum = measurement.eels.spectrum
            spectroscopy.eels = EELSProperty(
                resolution=instrument.eels.resolution,
                detector_type=instrument.eels.detector_type,
                min_energy=instrument.eels.min_energy,
                max_energy=instrument.eels.max_energy)



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, MEnum, JSON)
from nomad.datamodel.metainfo import simulation


m_package = Package(
    name='None',
    description='None')


class x_asr_parameters(MSection):

    m_def = Section(
        validate=False)

    x_asr_tmp_atoms = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_tmp_atoms_file = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_fmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_asr_calculator = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_asr_magstatecalculator = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_asr_rc = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_asr_d = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_asr_fsname = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_sc = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        ''')

    x_asr_dist_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class x_asr_code(MSection):

    m_def = Section(
        validate=False)

    x_asr_package = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_version = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_git_hash = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class x_asr_codes(MSection):

    m_def = Section(
        validate=False)

    x_asr_code = SubSection(
        sub_section=SectionProxy('x_asr_code'),
        repeats=True)


class x_asr_run_specification(MSection):

    m_def = Section(
        validate=False)

    x_asr_name = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_version = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_asr_uid = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_parameters = SubSection(
        sub_section=SectionProxy('x_asr_parameters'),
        repeats=False)

    x_asr_codes = SubSection(
        sub_section=SectionProxy('x_asr_codes'),
        repeats=False)


class x_asr_resources(MSection):

    m_def = Section(
        validate=False)

    x_asr_execution_start = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_asr_execution_end = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_asr_execution_duration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_asr_ncores = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')


class x_asr_dependency(MSection):

    m_def = Section(
        validate=False)

    x_asr_uid = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_revision = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class x_asr_dependencies(MSection):

    m_def = Section(
        validate=False)

    x_asr_dependency = SubSection(
        sub_section=SectionProxy('x_asr_dependency'),
        repeats=True)


class x_asr_metadata(MSection):

    m_def = Section(
        validate=False)

    x_asr_directory = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_created = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_asr_modified = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_asr_history = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_name = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_tags = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_revision = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_uid = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_asr_version = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_asr_resources = SubSection(
        sub_section=SectionProxy('x_asr_resources'),
        repeats=False)

    x_asr_dependencies = SubSection(
        sub_section=SectionProxy('x_asr_dependencies'),
        repeats=False)

    x_asr_metadata = SubSection(
        sub_section=SectionProxy('x_asr_metadata'),
        repeats=False)

    x_asr_run_specification = SubSection(
        sub_section=SectionProxy('x_asr_run_specification'),
        repeats=False)


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)
from nomad.datamodel.metainfo import simulation, workflow


m_package = Package()


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_openkim_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        build date as string
        ''')

    x_openkim_src_date = Quantity(
        type=str,
        shape=[],
        description='''
        date of last modification of the source as string
        ''')

    x_openkim_inserted_on = Quantity(
        type=str,
        shape=[],
        description='''
        title for the property
        ''')

    x_openkim_property_id = Quantity(
        type=str,
        shape=[],
        description='''
        unique ID of the property
        ''')

    x_openkim_property_title = Quantity(
        type=str,
        shape=[],
        description='''
        title for the property
        ''')

    x_openkim_property_description = Quantity(
        type=str,
        shape=[],
        description='''
        title for the property
        ''')

    x_openkim_instance_id = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        unique ID of the property
        ''')

    x_openkim_latest = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_openkim_meta = Quantity(
        type=JSON,
        shape=[],
        description='''
        openkim metadata'''
    )


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_openkim_short_name = Quantity(
        type=str,
        shape=[1],
        description='''
        short name defining the crystal
        ''')

    x_openkim_space_group = Quantity(
        type=str,
        shape=[1],
        description='''
        short name defining the crystal
        ''')


class Elastic(workflow.Elastic):

    m_def = Section(validate=False, extends_base_section=True)

    x_openkim_excess = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        total square numerical asymmetry of the calculated elastic constants
        ''')


class Phonon(workflow.Phonon):

    m_def = Section(validate=False, extends_base_section=True)

    x_openkim_wave_number = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints'],
        unit='1 / m',
        description='''
        wave numbers for each k-point
        ''')


class Workflow(workflow.Workflow):

    m_def = Section(validate=False, extends_base_section=True)

    x_openkim_property = Quantity(
        type=str,
        shape=[],
        description='''
        name of the property to be compared to nomad
        ''')

    x_openkim_nomad_rms_error = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        root mean square difference of the openkim data with respect to nomad
        ''')

    x_openkim_nomad_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        standard deviation of the nomad data
        ''')

    x_openkim_n_nomad_data = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of nomad entries with property corresponding to x_openkim_property
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import requests
from keycloak import KeycloakOpenID
import time

from nomad import config


class APIError(Exception): pass


def _call_requests(method, path: str, ssl: bool = True, *args, **kwargs):
    url = f'{config.client.url}/v1/{path}'
    return getattr(requests, method)(url, *args, **kwargs)


def get(*args, **kwargs):
    return _call_requests('get', *args, **kwargs)


def post(*args, **kwargs):
    return _call_requests('post', *args, **kwargs)


def put(*args, **kwargs):
    return _call_requests('put', *args, **kwargs)


def delete(*args, **kwargs):
    return _call_requests('delete', *args, **kwargs)


def url(path):
    ''' Returns the full NOMAD API url for the given api path. '''
    return f'{config.client.url}/v1/{path}'


class Auth(requests.auth.AuthBase):
    '''
    A request Auth class that can be used to authenticate in request callcs like this:

    .. code::

        requests.get('https://nomad-lab.eu/prod/rae/api/v1/entries', auth=Auth())

    This Auth will given username/password to acquire a token, it will keep and manage
    the token, and apply it to the headers requests.

    Arguments:
        user: Optional user name or email, default is take from ``config.client.user``
        password: Optional password, default is taken from ``config.client.password``
        from_api: If true, the necessary access token is acquired through the NOMAD api via basic auth
            and not via keycloak directly. Default is False. Not recommended, but might
            be useful, if keycloak can't be configured (e.g. during tests) or reached.
    '''
    def __init__(
            self, user: str = config.client.user,
            password: str = config.client.password,
            from_api: bool = False):
        self.user = user
        self._password = password
        self.from_api = from_api

        self.__oidc = KeycloakOpenID(
            server_url=config.keycloak.server_url,
            realm_name=config.keycloak.realm_name,
            client_id=config.keycloak.client_id)

        self._token = None

    def get_access_token_from_api(self):
        if self._token is None:
            response = requests.get(
                url('auth/token'),
                params=dict(username=self.user, password=self._password))

            if response.status_code != 200:
                response_json = response.json()
                raise APIError(
                    f'Could not acquire authentication token: '
                    f'{response_json.get("description") or response_json.get("detail") or "unknown reason"} '
                    f'({response_json.get("code", response.status_code)})')

            self._token = response.json()

    def get_access_token_from_keycloak(self):
        if self._token is None:
            self._token = self.__oidc.token(username=self.user, password=self._password)
            self._token['time'] = time.time()
        elif self._token['expires_in'] < int(time.time()) - self._token['time'] + 10:
            try:
                self._token = self.__oidc.refresh_token(self._token['refresh_token'])
                self._token['time'] = time.time()
            except Exception:
                self._token = self.__oidc.token(username=self.user, password=self._password)
                self._token['time'] = time.time()

    def __call__(self, request):
        if self.from_api:
            self.get_access_token_from_api()
        else:
            self.get_access_token_from_keycloak()

        request.headers['Authorization'] = f'Bearer {self._token["access_token"]}'
        return request



# Import from Log-Files Created by the PVD software by Roland Mainz
# for PVD Measurementfile v3, 3.0.0
# Version 0.1
# by Pascal Becker

import json
from datetime import datetime
import numpy as np
import pandas as pd


class Importer:
    def __init__(self):
        self.import_name = "pvd_p_evaporation"
        self.import_type = "processes"
        self.import_long_name = "PVD-P software log"
        self.supported_file_types = "Process logs (*.csv)"

    @staticmethod
    def get_unit(column_name: str) -> str:
        # Eurotherm values
        if column_name.endswith(' Aout'):
            return '%'
        if column_name.endswith(' Pv'):
            return '0.1 nmol cm$^{-1}$ s$^{-1}$'
        if column_name.endswith(' T') or column_name.endswith(' Tsp'):
            return '$^{\\circ}$C'

        # Inficon QCM values
        if column_name[:-1].endswith('Ch_Time_'):
            return 's'
        if column_name[:-1].endswith('C_Thik_'):
            return 'kA$^{\\circ}$'
        if column_name[:-1].endswith('Ratet1_'):
            return 'A$^{\\circ}$ s$^{-1}$'
        if column_name[:-1].endswith('Xlife_'):
            return '%'

        # Pressure
        if column_name[:-1].endswith(' Pressure'):
            return 'mbar'

        # none of the above (default)
        return ''

    def read(self, f, comment: str = None, user: str = None):
        header = '{"'
        while True:
            line = f.readline()
            if line[0] == '#':
                if ':' in line:
                    header += line[2:-1].replace(': ', '": "') + '", "'
            else:
                process_data = pd.read_csv(f, sep='\t')
                break

        header = header[:-3] + '}'
        header_dict = json.loads(header)
        header_dict['datetime'] = datetime.strptime(
            header_dict['Date'] + ' ' + header_dict['Time'], '%Y/%m/%d %H:%M:%S')

        evaporation_start_index = np.where(
            process_data['QCM1 SHTSRC_1'] + process_data['QCM1 SHTSRC_2'] + process_data['QCM2 SHTSRC_1'] + process_data['QCM2 SHTSRC_2'] > 0)[0][0]

        datetime_column = []
        process_time_column = []
        for process_time in process_data['Time']:
            datetime_column.append(datetime.strptime(header_dict['Date'] + ' ' + process_time, '%Y/%m/%d %H:%M:%S'))
        evaporation_start_time = datetime_column[evaporation_start_index]
        for process_time in datetime_column:
            if process_time < evaporation_start_time:
                process_time_column.append(-1 * (evaporation_start_time - process_time).seconds)
            else:
                process_time_column.append((process_time - evaporation_start_time).seconds)

        return process_time_column, process_data['Substrate PV'].to_numpy(),\
            process_data['Substrate TSP'].to_numpy(), process_data['Vacuum Pressure1'].to_numpy()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os.path
from typing import Tuple, List, Dict

from nomad import config
from nomad.datamodel import EntryArchive, EntryMetadata, results
from nomad.datamodel.context import Context

from .parser import MissingParser, BrokenParser, Parser, ArchiveParser, MatchingParserInterface
from .artificial import EmptyParser, GenerateRandomParser, TemplateParser, ChaosParser


try:
    # these packages are not available without parsing extra, which is ok, if the
    # parsers are only initialized to load their metainfo definitions
    import magic
    import gzip
    import bz2
    import lzma

    _compressions = {
        b'\x1f\x8b\x08': ('gz', gzip.open),
        b'\x42\x5a\x68': ('bz2', bz2.open),
        b'\xfd\x37\x7a': ('xz', lzma.open)
    }

    encoding_magic = magic.Magic(mime_encoding=True)

except ImportError:
    pass


def match_parser(mainfile_path: str, strict=True, parser_name: str = None) -> Tuple[Parser, List[str]]:
    '''
    Performs parser matching. This means it take the given mainfile and potentially
    opens it with the given callback and tries to identify a parser that can parse
    the file.

    This is determined by filename (e.g. *.out), mime type (e.g. text/*, application/xml),
    and beginning file contents.

    Arguments:
        mainfile_path: Path to the mainfile
        strict: Only match strict parsers, e.g. no artificial parsers for missing or empty entries.
        parser_name: Optional, to force the matching to test only a specific parser

    Returns:
        A tuple (`parser`, `mainfile_keys`). The `parser` is the matched parser, and
        `mainfile_keys` defines the keys to use for child entries, if any. If there are
        no child entries, `mainfile_keys` will be None. If no parser matches, we return
        (None, None).
    '''
    mainfile = os.path.basename(mainfile_path)
    if mainfile.startswith('.') or mainfile.startswith('~'):
        return None, None

    with open(mainfile_path, 'rb') as f:
        compression, open_compressed = _compressions.get(f.read(3), (None, open))

    with open_compressed(mainfile_path, 'rb') as cf:  # type: ignore
        buffer = cf.read(config.parser_matching_size)

    mime_type = magic.from_buffer(buffer, mime=True)

    decoded_buffer = None
    encoding = None
    try:  # Try to open the file as a string for regex matching.
        decoded_buffer = buffer.decode('utf-8')
    except UnicodeDecodeError:
        # This file is either binary or has wrong encoding
        encoding = encoding_magic.from_buffer(buffer)

        if config.services.force_raw_file_decoding:
            encoding = 'iso-8859-1'

        if encoding in ['iso-8859-1']:
            try:
                decoded_buffer = buffer.decode(encoding)
            except Exception:
                pass
    if parser_name:
        parser = parser_dict.get(parser_name)
        assert parser is not None, f'parser by the name `{parser_name}` does not exist'
        parsers_to_check = [parser]
    else:
        parsers_to_check = parsers
    for parser in parsers_to_check:
        if strict and isinstance(parser, (MissingParser, EmptyParser)):
            continue

        match_result = parser.is_mainfile(mainfile_path, mime_type, buffer, decoded_buffer, compression)
        if match_result:
            if type(match_result) == set:
                assert parser.creates_children, 'Illegal return value - parser does not specify `creates_children`'
                for mainfile_key in match_result:  # type: ignore
                    assert mainfile_key and type(mainfile_key) == str, (
                        f'Child keys must be strings, got {type(mainfile_key)}')
                mainfile_keys = sorted(match_result)  # type: ignore
            else:
                mainfile_keys = None

            # potentially convert the file
            if encoding in ['iso-8859-1']:
                try:
                    with open(mainfile_path, 'rb') as binary_file:
                        content = binary_file.read().decode(encoding)
                except Exception:
                    pass
                else:
                    with open(mainfile_path, 'wt') as text_file:
                        text_file.write(content)

            # TODO: deal with multiple possible parser specs
            return parser, mainfile_keys

    return None, None


class ParserContext(Context):
    def __init__(self, mainfile_dir):
        self._mainfile_dir = mainfile_dir

    def raw_file(self, path, *args, **kwargs):
        return open(os.path.join(self._mainfile_dir, path), *args, **kwargs)


def run_parser(
        mainfile_path: str, parser: Parser, mainfile_keys: List[str] = None, logger=None) -> List[EntryArchive]:
    '''
    Parses a file, given the path, the parser, and mainfile_keys, as returned by
    :func:`match_parser`, and returns the resulting EntryArchive objects. Parsers that have
    `create_children == False` (the most common case) will generate a list with a single entry,
    for parsers that create children the list will consist of the main entry followed by the
    child entries. The returned archive objects will have minimal metadata.
    '''
    entry_archive = EntryArchive(m_context=ParserContext(os.path.dirname(mainfile_path)))
    metadata = entry_archive.m_create(EntryMetadata)
    metadata.mainfile = mainfile_path
    entry_archives = [entry_archive]
    if mainfile_keys:
        child_archives = {}
        for mainfile_key in mainfile_keys:
            child_archive = EntryArchive()
            child_metadata = child_archive.m_create(EntryMetadata)
            child_metadata.mainfile = mainfile_path
            child_metadata.mainfile_key = mainfile_key
            child_archives[mainfile_key] = child_archive
            entry_archives.append(child_archive)
        kwargs = dict(child_archives=child_archives)
    else:
        kwargs = {}

    cwd = os.getcwd()
    try:
        mainfile_path = os.path.abspath(mainfile_path)
        os.chdir(os.path.abspath(os.path.dirname(mainfile_path)))
        parser.parse(mainfile_path, entry_archive, logger=logger, **kwargs)
    except Exception as e:
        if logger:
            logger.error('parsing was not successful', exc_info=e)
        raise e
    finally:
        os.chdir(cwd)
    for entry_archive in entry_archives:
        if entry_archive.metadata.domain is None:
            entry_archive.metadata.domain = parser.domain
    return entry_archives


parsers = [
    GenerateRandomParser(),
    TemplateParser(),
    ChaosParser(),
    MatchingParserInterface(
        'electronicparsers.AbinitParser',
        name='parsers/abinit', code_name='ABINIT', code_homepage='https://www.abinit.org/',
        mainfile_contents_re=(r'^\n*\.Version\s*[0-9.]*\s*of ABINIT\s*')
    ),
    MatchingParserInterface(
        'electronicparsers.ATKParser',
        name='parsers/atk', code_name='AtomistixToolKit',
        code_homepage='https://www.synopsys.com/silicon/quantumatk.html',
        mainfile_name_re=r'^.*\.nc', mainfile_mime_re=r'application/octet-stream'
    ),
    MatchingParserInterface(
        'electronicparsers.BandParser',
        name='parsers/band', code_name='BAND',
        code_homepage='https://www.scm.com/product/band_periodicdft/',
        mainfile_contents_re=r' +\* +Amsterdam Density Functional +\(ADF\)'
    ),
    MatchingParserInterface(
        'electronicparsers.BigDFTParser',
        name='parsers/bigdft', code_name='BigDFT', code_homepage='http://bigdft.org/',
        mainfile_contents_re=(
            # r'__________________________________ A fast and precise DFT wavelet code\s*'
            # r'\|     \|     \|     \|     \|     \|\s*'
            # r'\|     \|     \|     \|     \|     \|      BBBB         i       gggggg\s*'
            # r'\|_____\|_____\|_____\|_____\|_____\|     B    B               g\s*'
            # r'\|     \|  :  \|  :  \|     \|     \|    B     B        i     g\s*'
            # r'\|     \|-0\+--\|-0\+--\|     \|     \|    B    B         i     g        g\s*'
            r'\|_____\|__:__\|__:__\|_____\|_____\|___ BBBBB          i     g         g\s*'
            # r'\|  :  \|     \|     \|  :  \|     \|    B    B         i     g         g\s*'
            # r'\|--\+0-\|     \|     \|-0\+--\|     \|    B     B     iiii     g         g\s*'
            # r'\|__:__\|_____\|_____\|__:__\|_____\|    B     B        i      g        g\s*'
            # r'\|     \|  :  \|  :  \|     \|     \|    B BBBB        i        g      g\s*'
            # r'\|     \|-0\+--\|-0\+--\|     \|     \|    B        iiiii          gggggg\s*'
            # r'\|_____\|__:__\|__:__\|_____\|_____\|__BBBBB\s*'
            # r'\|     \|     \|     \|  :  \|     \|                           TTTTTTTTT\s*'
            # r'\|     \|     \|     \|--\+0-\|     \|  DDDDDD          FFFFF        T\s*'
            # r'\|_____\|_____\|_____\|__:__\|_____\| D      D        F        TTTT T\s*'
            # r'\|     \|     \|     \|  :  \|     \|D        D      F        T     T\s*'
            # r'\|     \|     \|     \|--\+0-\|     \|D         D     FFFF     T     T\s*'
            # r'\|_____\|_____\|_____\|__:__\|_____\|D___      D     F         T    T\s*'
            # r'\|     \|     \|  :  \|     \|     \|D         D     F          TTTTT\s*'
            # r'\|     \|     \|--\+0-\|     \|     \| D        D     F         T    T\s*'
            # r'\|_____\|_____\|__:__\|_____\|_____\|          D     F        T     T\s*'
            # r'\|     \|     \|     \|     \|     \|         D               T    T\s*'
            # r'\|     \|     \|     \|     \|     \|   DDDDDD       F         TTTT\s*'
            # r'\|_____\|_____\|_____\|_____\|_____\|______                    www\.bigdft\.org'
        )
    ),
    MatchingParserInterface(
        'electronicparsers.CastepParser',
        name='parsers/castep', code_name='CASTEP', code_homepage='http://www.castep.org/',
        mainfile_contents_re=(r'\s\|\s*CCC\s*AA\s*SSS\s*TTTTT\s*EEEEE\s*PPPP\s*\|\s*')
    ),
    MatchingParserInterface(
        'electronicparsers.CharmmParser',
        name='parsers/charmm', code_name='Charmm', domain='dft',
        mainfile_contents_re=r'\s*Chemistry\s*at\s*HARvard\s*Macromolecular\s*Mechanics\s*',
        mainfile_mime_re=r'text/.*'
    ),
    MatchingParserInterface(
        'electronicparsers.CP2KParser',
        name='parsers/cp2k', code_name='CP2K', code_homepage='https://www.cp2k.org/',
        mainfile_contents_re=(
            r'\*\*\*\* \*\*\*\* \*\*\*\*\*\*  \*\*  PROGRAM STARTED AT\s.*\n'
            r' \*\*\*\*\* \*\* \*\*\*  \*\*\* \*\*   PROGRAM STARTED ON\s*.*\n'
            r' \*\*    \*\*\*\*   \*\*\*\*\*\*    PROGRAM STARTED BY .*\n'
            r' \*\*\*\*\* \*\*    \*\* \*\* \*\*   PROGRAM PROCESS ID .*\n'
            r'  \*\*\*\* \*\*  \*\*\*\*\*\*\*  \*\*  PROGRAM STARTED IN .*\n')
    ),
    MatchingParserInterface(
        'electronicparsers.CPMDParser',
        name='parsers/cpmd', code_name='CPMD',
        code_homepage='https://www.lcrc.anl.gov/for-users/software/available-software/cpmd/',
        mainfile_contents_re=(r'\*\*\*       \*\*   \*\*\*  \*\* \*\*\*\* \*\*  \*\*   \*\*\*')
    ),
    MatchingParserInterface(
        'electronicparsers.CrystalParser',
        name='parsers/crystal',
        code_name='Crystal',
        code_homepage='https://www.crystal.unito.it/',
        mainfile_contents_re=(
            fr'(\r?\n \*\s+CRYSTAL[\d]+\s+\*\r?\n \*\s*[a-zA-Z]+ : \d+[\.\d+]*)')
    ),
    MatchingParserInterface(
        'electronicparsers.Dmol3Parser',
        name='parsers/dmol', code_name='DMol3',
        code_homepage='http://dmol3.web.psi.ch/dmol3.html', domain='dft',
        mainfile_name_re=r'.*\.outmol',
        mainfile_contents_re=r'Materials Studio DMol\^3'
    ),
    MatchingParserInterface(
        'electronicparsers.ElkParser',
        name='parsers/elk', code_name='elk', code_homepage='http://elk.sourceforge.net/',
        mainfile_contents_re=r'\| Elk version [0-9.a-zA-Z]+ started \|'
    ),
    MatchingParserInterface(
        'electronicparsers.ExcitingParser',
        name='parsers/exciting', code_name='exciting', code_homepage='http://exciting-code.org/',
        mainfile_name_re=r'^.*.OUT(\.[^/]*)?$', mainfile_contents_re=(r'EXCITING.*started')
    ),
    MatchingParserInterface(
        'electronicparsers.FHIAimsParser',
        name='parsers/fhi-aims', code_name='FHI-aims',
        code_homepage='https://aimsclub.fhi-berlin.mpg.de/',
        mainfile_contents_re=(
            r'^(.*\n)*'
            r'?\s*Invoking FHI-aims \.\.\.')
    ),
    MatchingParserInterface(
        'electronicparsers.FleurParser',
        name='parsers/fleur', code_name='fleur',
        code_homepage='https://www.flapw.de/', domain='dft',
        mainfile_contents_re=r'This output is generated by fleur.'
    ),
    MatchingParserInterface(
        'electronicparsers.FploParser',
        name='parsers/fplo', code_name='fplo', domain='dft',
        mainfile_contents_re=r'\s*\|\s*FULL-POTENTIAL LOCAL-ORBITAL MINIMUM BASIS BANDSTRUCTURE CODE\s*\|\s*',
        mainfile_mime_re=r'text/.*'
    ),
    MatchingParserInterface(
        'electronicparsers.GamessParser',
        name='parsers/gamess', code_name='GAMESS',
        code_homepage='https://www.msg.chem.iastate.edu/gamess/versions.html',
        mainfile_contents_re=(
            r'\s*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
            r'\s*\*\s*GAMESS VERSION =\s*(.*)\*\s*'
            r'\s*\*\s*FROM IOWA STATE UNIVERSITY\s*\*\s*')
    ),
    MatchingParserInterface(
        'electronicparsers.GaussianParser',
        name='parsers/gaussian', code_name='Gaussian', code_homepage='http://gaussian.com/',
        mainfile_mime_re=r'.*', mainfile_contents_re=(
            r'\s*Cite this work as:'
            r'\s*Gaussian [0-9]+, Revision [A-Za-z0-9\.]*,')
    ),
    MatchingParserInterface(
        'electronicparsers.GPAWParser',
        name='parsers/gpaw', code_name='GPAW', code_homepage='https://wiki.fysik.dtu.dk/gpaw/',
        mainfile_name_re=(r'^.*\.(gpw2|gpw)$'),
        mainfile_mime_re=r'application/(x-tar|octet-stream)'
    ),
    MatchingParserInterface(
        'electronicparsers.MolcasParser',
        name='parsers/molcas', code_name='MOLCAS', code_homepage='http://www.molcas.org/',
        domain='dft', mainfile_contents_re=r'M O L C A S'
    ),
    MatchingParserInterface(
        'electronicparsers.NWChemParser',
        name='parsers/nwchem', code_name='NWChem', code_homepage='http://www.nwchem-sw.org/',
        mainfile_contents_re=(
            r'Northwest Computational Chemistry Package \(NWChem\) (\d+\.)+\d+')
    ),
    MatchingParserInterface(
        'electronicparsers.OctopusParser',
        name='parsers/octopus', code_name='Octopus', code_homepage='https://octopus-code.org/',
        mainfile_contents_re=(r'\|0\) ~ \(0\) \|')
    ),
    MatchingParserInterface(
        'electronicparsers.OnetepParser',
        name='parsers/onetep', code_name='ONETEP', code_homepage='https://www.onetep.org/',
        domain='dft', mainfile_contents_re=r'####### #     # ####### ####### ####### ######'
    ),
    MatchingParserInterface(
        'electronicparsers.OrcaParser',
        name='parsers/orca', code_name='ORCA', code_homepage='https://orcaforum.kofo.mpg.de/',
        mainfile_contents_re=(
            r'\s+\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
            r'\s+\* O   R   C   A \*\s*'
            r'\s+\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
            r'\s*'
            r'\s*--- An Ab Initio, DFT and Semiempirical electronic structure package ---\s*')
    ),
    MatchingParserInterface(
        'electronicparsers.Psi4Parser',
        name='parsers/psi4', code_name='Psi4', code_homepage='https://psicode.org/',
        mainfile_contents_re=(r'Psi4: An Open-Source Ab Initio Electronic Structure Package')
    ),
    MatchingParserInterface(
        'electronicparsers.QBallParser',
        name='parsers/qball',
        code_name='qball',
        mainfile_contents_re='qball',
        supported_compressions=["gz", "bz2", "xz"]
    ),
    MatchingParserInterface(
        'electronicparsers.QboxParser',
        name='parsers/qbox', code_name='qbox', code_homepage='http://qboxcode.org/',
        domain='dft', mainfile_mime_re=r'(application/xml)|(text/.*)',
        mainfile_contents_re=(r'http://qboxcode.org')
    ),
    MatchingParserInterface(
        'electronicparsers.QuantumEspressoParser',
        name='parsers/quantumespresso', code_name='Quantum Espresso',
        code_homepage='https://www.quantum-espresso.org/',
        mainfile_contents_re=(r'(Program PWSCF.*starts)|(Current dimensions of program PWSCF are)')
    ),
    MatchingParserInterface(
        'electronicparsers.SiestaParser',
        name='parsers/siesta', code_name='Siesta',
        code_homepage='https://departments.icmab.es/leem/siesta/',
        mainfile_contents_re=(
            r'(Siesta Version: siesta-|SIESTA [0-9]\.[0-9]\.[0-9])|'
            r'(\*\s*WELCOME TO SIESTA\s*\*)')
    ),
    MatchingParserInterface(
        'electronicparsers.TurbomoleParser',
        name='parsers/turbomole', code_name='turbomole',
        code_homepage='https://www.turbomole.org/',
        mainfile_contents_re=(r'Copyright \(C\) [0-9]+ TURBOMOLE GmbH, Karlsruhe')
    ),
    MatchingParserInterface(
        'electronicparsers.VASPParser',
        name='parsers/vasp', code_name='VASP', code_homepage='https://www.vasp.at/',
        mainfile_mime_re=r'(application/.*)|(text/.*)',
        mainfile_name_re=r'.*[^/]*xml[^/]*',  # only the alternative mainfile name should match
        mainfile_contents_re=(
            r'^\s*<\?xml version="1\.0" encoding="ISO-8859-1"\?>\s*'
            r'?\s*<modeling>'
            r'?\s*<generator>'
            r'?\s*<i name="program" type="string">\s*vasp\s*</i>'
            r'?|^\svasp[\.\d]+.+?(?:\(build|complex)[\s\S]+?executed on'),
        supported_compressions=['gz', 'bz2', 'xz'], mainfile_alternative=True
    ),
    MatchingParserInterface(
        'electronicparsers.Wien2kParser',
        name='parsers/wien2k', code_name='WIEN2k', code_homepage='http://www.wien2k.at/',
        mainfile_name_re=r'.*\.scf$',
        mainfile_alternative=True,
        mainfile_contents_re=r'\s*---------\s*:ITE[0-9]+:\s*[0-9]+\.\s*ITERATION\s*---------'
    ),
    MatchingParserInterface(
        'electronicparsers.YamboParser',
        name='parsers/yambo', code_name='YAMBO', code_homepage='https://yambo-code.org/',
        mainfile_contents_re=(r'Build.+\s+http://www\.yambo-code\.org')
    ),
    MatchingParserInterface(
        'atomisticparsers.AmberParser',
        name='parsers/amber', code_name='Amber', domain='dft',
        mainfile_contents_re=r'\s*Amber\s[0-9]+\s[A-Z]+\s*[0-9]+'
    ),
    MatchingParserInterface(
        'atomisticparsers.AsapParser',
        name='parsers/asap', code_name='ASAP', domain='dft',
        mainfile_name_re=r'.*.traj$', mainfile_mime_re=r'application/octet-stream'
    ),
    MatchingParserInterface(
        'atomisticparsers.DFTBPlusParser',
        name='parsers/dftbplus', code_name='DFTB+', domain='dft',
        mainfile_contents_re=r'\|  DFTB\+',
        mainfile_mime_re=r'text/.*'
    ),
    MatchingParserInterface(
        'atomisticparsers.DLPolyParser',
        name='parsers/dl-poly', code_name='DL_POLY',
        code_homepage='https://www.scd.stfc.ac.uk/Pages/DL_POLY.aspx',
        mainfile_contents_re=(r'\*\* DL_POLY \*\*'),
    ),
    MatchingParserInterface(
        'atomisticparsers.GromacsParser',
        name='parsers/gromacs', code_name='Gromacs', code_homepage='http://www.gromacs.org/',
        domain='dft', mainfile_contents_re=r'gmx mdrun, (VERSION|version)'
    ),
    MatchingParserInterface(
        'atomisticparsers.GromosParser',
        name='parsers/gromos', code_name='Gromos', domain='dft',
        mainfile_contents_re=r'Bugreports to http://www.gromos.net'
    ),
    MatchingParserInterface(
        'atomisticparsers.GulpParser',
        name='parsers/gulp', code_name='gulp', code_homepage='http://gulp.curtin.edu.au/gulp/',
        mainfile_contents_re=(
            r'\s*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*'
            r'\*\*\*\*\*\*\*\*\*\*\*\*\*\s*'
            r'\s*\*\s*GENERAL UTILITY LATTICE PROGRAM\s*\*\s*'),
    ),
    MatchingParserInterface(
        'atomisticparsers.LammpsParser',
        name='parsers/lammps', code_name='LAMMPS', code_homepage='https://lammps.sandia.gov/',
        domain='dft', mainfile_contents_re=r'^LAMMPS'
    ),
    MatchingParserInterface(
        'atomisticparsers.LibAtomsParser',
        name='parsers/lib-atoms', code_name='libAtoms', code_homepage='https://libatoms.github.io/',
        mainfile_contents_re=(r'\s*<GAP_params\s'),
    ),
    MatchingParserInterface(
        'atomisticparsers.MopacParser',
        name='parsers/mopac', code_name='MOPAC', domain='dft',
        mainfile_contents_re=r'\s*\*\*\s*MOPAC\s*([0-9a-zA-Z]*)\s*\*\*\s*',
        mainfile_mime_re=r'text/.*',
    ),
    MatchingParserInterface(
        'atomisticparsers.NAMDParser',
        name='parsers/namd', code_name='Namd', domain='dft',
        mainfile_contents_re=r'\s*Info:\s*NAMD\s*[0-9.]+\s*for\s*',
        mainfile_mime_re=r'text/.*',
    ),
    MatchingParserInterface(
        'atomisticparsers.OpenKIMParser',
        name='parsers/openkim', code_name='OpenKIM', domain='dft',
        mainfile_mime_re=r'(application/json)|(text/.*)',
        mainfile_contents_re=r'openkim|OPENKIM|OpenKIM'
    ),
    MatchingParserInterface(
        'atomisticparsers.TinkerParser',
        name='parsers/tinker', code_name='TINKER', domain='dft',
        mainfile_contents_re=r'TINKER  ---  Software Tools for Molecular Design'
    ),
    MatchingParserInterface(
        'workflowparsers.AFLOWParser',
        name='parsers/aflow', code_name='AFlow', code_homepage='http://www.aflowlib.org/',
        mainfile_mime_re=r'(application/json)|(text/.*)',
        mainfile_name_re=r'.*aflowlib\.json.*',  # only the alternative mainfile name should match
        mainfile_contents_re=(
            r"^\s*\[AFLOW\] \*+"
            r"\s*\[AFLOW\]"
            r"\s*\[AFLOW\]                     .o.        .o88o. oooo"
            r"\s*\[AFLOW\]                    .888.       888 `` `888"
            r"\s*\[AFLOW\]                   .8'888.     o888oo   888   .ooooo.  oooo oooo    ooo"
            r"\s*\[AFLOW\]                  .8' `888.     888     888  d88' `88b  `88. `88.  .8'"
            r"\s*\[AFLOW\]                 .88ooo8888.    888     888  888   888   `88..]88..8'"
            r"\s*\[AFLOW\]                .8'     `888.   888     888  888   888    `888'`888'"
            r"\s*\[AFLOW\]               o88o     o8888o o888o   o888o `Y8bod8P'     `8'  `8'  .in"
            r"|^\s*\{\"aurl\"\:\"aflowlib\.duke\.edu\:AFLOWDATA"),
        supported_compressions=['gz', 'bz2', 'xz'], mainfile_alternative=True
    ),
    MatchingParserInterface(
        'workflowparsers.ASRParser',
        name='parsers/asr', code_name='ASR',
        code_homepage='https://asr.readthedocs.io/en/latest/index.html',
        mainfile_mime_re=r'(application/json)|(text/.*)',
        mainfile_name_re=r'.*archive_.*\.json',
        mainfile_contents_re=(r'"name": "ASR"')
    ),
    MatchingParserInterface(
        'workflowparsers.ElasticParser',
        name='parsers/elastic', code_name='elastic', code_homepage='http://exciting-code.org/elastic',
        mainfile_contents_re=r'\s*Order of elastic constants\s*=\s*[0-9]+\s*',
        mainfile_name_re=(r'.*/INFO_ElaStic')
    ),
    MatchingParserInterface(
        'workflowparsers.FHIVibesParser',
        name='parsers/fhi-vibes', code_name='FHI-vibes',
        code_homepage='https://vibes.fhi-berlin.mpg.de/',
        mainfile_name_re=(r'^.*\.(nc)$'), mainfile_mime_re=r'(application/x-hdf)',
        mainfile_binary_header_re=br'^\x89HDF'
    ),
    MatchingParserInterface(
        'workflowparsers.LobsterParser',
        name='parsers/lobster', code_name='LOBSTER',
        code_homepage='http://schmeling.ac.rwth-aachen.de/cohp/',
        mainfile_name_re=r'.*lobsterout$',
        mainfile_contents_re=(r'^LOBSTER\s*v[\d\.]+.*'),
    ),
    MatchingParserInterface(
        'workflowparsers.MPParser',
        name='parsers/mp', code_name='MaterialsProject',
        code_homepage='https://materialsproject.org',
        mainfile_mime_re=r'(application/json)|(text/.*)',
        mainfile_name_re=r'.*mp.+materials\.json',
        mainfile_contents_re=(r'"pymatgen_version":')
    ),
    MatchingParserInterface(
        'workflowparsers.PhonopyParser',
        name='parsers/phonopy', code_name='Phonopy', code_homepage='https://phonopy.github.io/phonopy/',
        mainfile_name_re=(r'(.*/phonopy-FHI-aims-displacement-0*1/control.in$)|(.*/phon.+yaml)')
    ),
    MatchingParserInterface(
        'eelsdbparser.EELSDBParser',
        name='parsers/eels', code_name='eels', code_homepage='https://eelsdb.eu/',
        domain='ems',
        mainfile_mime_re=r'application/json',
        mainfile_contents_re=(r'https://eelsdb.eu/spectra')
    ),
    MatchingParserInterface(
        'nexusparser.NexusParser',
        name='parsers/nexus', code_name='NEXUS', code_homepage='https://www.nexus.eu/',
        mainfile_mime_re=r'(application/.*)|(text/.*)',
        mainfile_name_re=(r'.*\.nxs'),
        supported_compressions=['gz', 'bz2', 'xz']
    ),
    ArchiveParser()
]

empty_parsers = [
    EmptyParser(
        name='missing/octopus', code_name='Octopus', code_homepage='https://octopus-code.org/',
        domain='dft',
        mainfile_name_re=r'(inp)|(.*/inp)'
    ),
    EmptyParser(
        name='missing/crystal', code_name='Crystal', code_homepage='https://www.crystal.unito.it/index.php',
        domain='dft',
        mainfile_name_re=r'.*\.cryst\.out'
    ),
    EmptyParser(
        name='missing/wien2k', code_name='WIEN2k', code_homepage='http://www.wien2k.at/',
        domain='dft',
        mainfile_name_re=r'.*\.scf'
    ),
    EmptyParser(
        name='missing/fhi-aims', code_name='FHI-aims', code_homepage='https://aimsclub.fhi-berlin.mpg.de/',
        domain='dft',
        mainfile_name_re=r'.*\.fhiaims'
    )
]

if config.use_empty_parsers:
    # There are some entries with PIDs that have mainfiles which do not match what
    # the actual parsers expect. We use the EmptyParser to produce placeholder entries
    # to keep the PIDs. These parsers will not match for new, non migrated data.
    parsers.extend(empty_parsers)

parsers.append(BrokenParser())

''' Instantiation and constructor based config of all parsers. '''

parser_dict: Dict[str, Parser] = {parser.name: parser for parser in parsers + empty_parsers}  # type: ignore
''' A dict to access parsers by name. Usually 'parsers/<...>', e.g. 'parsers/vasp'. '''

# renamed parsers
parser_dict['parsers/vaspoutcar'] = parser_dict['parsers/vasp']
parser_dict['parser/broken'] = parser_dict['parsers/broken']
parser_dict['parser/fleur'] = parser_dict['parsers/fleur']
parser_dict['parser/molcas'] = parser_dict['parsers/molcas']
parser_dict['parser/octopus'] = parser_dict['parsers/octopus']
parser_dict['parser/onetep'] = parser_dict['parsers/onetep']

# register code names as possible statistic value to the dft datamodel
code_names = []
for parser in parsers:
    if parser.domain == 'dft' and \
            getattr(parser, 'code_name', None) is not None and \
            getattr(parser, 'code_name') != 'currupted mainfile' and \
            getattr(parser, 'code_name') != 'Template':
        code_names.append(getattr(parser, 'code_name'))
code_names = sorted(set(code_names), key=lambda code_name: code_name.lower())
results.Simulation.program_name.a_elasticsearch[0].values = code_names + [
    config.services.unavailable_value, config.services.not_processed_value]


def import_all_parsers():
    '''
    Imports all the parsers. This will instantiate all parser metainfo as a side
    effect.
    '''
    for parser in parsers:
        if isinstance(parser, MatchingParserInterface):
            parser.import_parser_class()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This elasticsearch extension for the Metainfo allows to define how quantities are
added to Elasticsearch indices.

This extension supports two search indices: ``entry_index`` and ``material_index``.
There are three different types of "searchable documents": ``entry_type``, ``material_type``,
``material_entry_type``. Entry documents are indexed in the entry index; material documents
in the material index. The material entry documents are nested documents in material documents.

The document types are subsets of the metainfo schema; documents have the exact same
structure as archives, but with only some of the quantities. Which quantities are in these
documents can be defined in the metainfo by using the :class:`Elasticsearch` annotation on
quantity definitions.

Entry and material entry documents start with the metainfo entry root section.
Material documents start with the ``results.material`` sub-section. Nested material entry
documents are placed under the ``entries`` key within a material document. This is the only
exception, where the material document structure deviates from the metainfo/archive structure.

A quantity can appear in multiple document types. All indexed quantities
appear by default in entry documents. If specified quantities are also put in either
the material document or a nested material entry document within the material document.
The material quantities describe the material itself
(e.g. formula, elements, system type, symmetries). These quantities are always in all
entries of the same material. Material entry quantities describe individual results
and metadata that are contributed by the entries of this material (e.g. published, embargo,
band gap, available properties). The values contributed by different entries of the same
material may vary.

Here is a small metainfo example:

.. code-block:: python

    class Entry(MSection):

        entry_id = Quantity(
            type=str,
            a_elasticsearch=Elasticsearch(material_entry_type))

        upload_create_time = Quantity(
            type=Datetime,
            a_elasticsearch=Elasticsearch())

        results = SubSection(sub_section=Results.m_def, a_elasticsearch=Elasticsearch())


    class Results(MSection):

        material = SubSection(sub_section=Material.m_def, a_elasticsearch=Elasticsearch())
        properties = SubSection(sub_section=Properties.m_def, a_elasticsearch=Elasticsearch())


    class Material(MSection):

        material_id = Quantity(
            type=str,
            a_elasticsearch=Elasticsearch(material_type))

        formula = Quantity(
            type=str,
            a_elasticsearch=[
                Elasticsearch(material_type),
                Elasticsearch(material_type, field='text', mapping='text')])


    class Properties(MSection):

        available_properties = Quantity(
            type=str, shape=['*'],
            a_elasticsearch=Elasticsearch(material_entry_type))

        band_gap = Quantity(
            type=float, unit='J',
            a_elasticsearch=Elasticsearch(material_entry_type))


The resulting indices with a single entry in them would look like this. Entry index:

.. code-block:: json

    [
        {
            "entry_id": "de54f1",
            "upload_create_time": "2021-02-01 01:23:12",
            "results": {
                "material": {
                    "material_id": "23a8bf",
                    "formula": "H2O"
                },
                "properties": {
                    "available_properties": ["dos", "bs", "band_gap", "energy_total_0"],
                    "band_gap": 0.283e-12
                }
            }
        }
    ]


And material index:

.. code-block:: json

    [
        {
            "material_id": "23a8bf",
            "formula": "H2O"
            "entries": [
                {
                    "entry_id": "de54f1",
                    "results": {
                        "properties": {
                            "available_properties": ["dos", "bs", "band_gap", "energy_total_0"],
                            "band_gap": 0.283e-12
                        }
                    }
                }
            ]
        }
    ]

You can freely define sub-sections and quantities. The only fixed structures that are
required from the metainfo are:
- the root section has an ``entry_id``
- materials are placed in ``results.material``
- the ``results.material`` sub-section has a ``material_id``
- the ``results.material`` sub-section has no property called ``entries``

This extension resolves references during indexing and basically treats referenced
sub-sections as if they were direct sub-sections.

.. autofunction:: index_entry
.. autofunction:: index_entries
.. autofunction:: create_indices


.. autoclass:: Elasticsearch
.. autoclass:: DocumentType
.. autoclass:: Index
'''


from typing import Union, Any, Dict, cast, Set, List, Callable, Tuple, DefaultDict
from collections import defaultdict
import numpy as np
import re

from nomad import config, utils

from .metainfo import (
    MSectionBound, Section, Quantity, MSection, MEnum, Datetime, Reference, DefinitionAnnotation,
    Definition, QuantityReference, Unit)


class DocumentType():
    '''
    DocumentType allows to create Elasticsearch index mappings and documents based on
    Metainfo definitions and instances. Genrally this class should not be used outside
    the elasticsearch_extension module.

    Attributes:
        root_section_def: The section definition that serves as the root for all documents.
            mapping: The elasticsearch mapping definition.
        indexed_properties: All definitions (quantities and sub sections) that are covered
            by documents of this type.
        quantities: All elasticsearch quantities that in documents of this type. A dictionary
            with full qualified name as key and :class:`Elasticsearch` annotations as
            values.
        metrics: All metrics in this document type. A dictionary with metric names as
            keys and tuples of elasticsearch metric aggregation and respective
            :class:`Elasticsearch` metainfo annotation as values.
        id_field: The quantity (and elasticsearch field) name that is used as unique
            identifier for this type of documents.
    '''
    def __init__(self, name: str, id_field: str):
        self.name = name
        self.id_field = id_field
        self.root_section_def = None
        self.mapping: Dict[str, Any] = None
        self.indexed_properties: Set[Definition] = set()
        self.nested_object_keys: List[str] = []
        self.nested_sections: List[SearchQuantity] = []
        self.quantities: Dict[str, SearchQuantity] = {}
        self.suggestions: Dict[str, Elasticsearch] = {}
        self.metrics: Dict[str, Tuple[str, SearchQuantity]] = {}

    def _reset(self):
        self.indexed_properties.clear()
        self.nested_object_keys.clear()
        self.nested_sections.clear()
        self.quantities.clear()
        self.metrics.clear()

    def create_index_doc(self, root: MSection):
        '''
        Creates an indexable document from the given archive.
        '''
        suggestions: DefaultDict = defaultdict(list)

        def transform(quantity, section, value, path):
            '''
            Custom transform function that possibly transforms the indexed
            values and also gathers the suggestion values for later storage.
            '''
            elasticsearch_annotations = quantity.m_get_annotations(Elasticsearch, as_list=True)
            for elasticsearch_annotation in elasticsearch_annotations:
                if elasticsearch_annotation.field is None:
                    if elasticsearch_annotation.suggestion:
                        # The suggestions may have a different doc_type: we
                        # don't serialize them if the doc types don't match.
                        if self != entry_type and elasticsearch_annotation.doc_type != self:
                            continue

                        # The suggestion values are saved into a temporary
                        # dictionary. The actual path of the data in the
                        # metainfo is used as a key. The suggestions will also
                        # take into account any given variants of the input.
                        transform_function = elasticsearch_annotation.value
                        variants = elasticsearch_annotation.variants
                        if transform_function is not None:
                            if variants:
                                suggestion_value = []
                                for variant in variants(value):
                                    suggestion_value.extend(transform_function(variant))
                                suggestion_value = list(set(suggestion_value))
                            else:
                                suggestion_value = transform_function(value)
                        else:
                            suggestion_value = value
                        section_path = section.m_path()[len(root.m_path()):]
                        name = elasticsearch_annotation.property_name
                        if path:
                            suggestion_path = f"{section_path}/{path}/{name}"
                        else:
                            suggestion_path = f"{section_path}/{name}"
                        suggestions[suggestion_path].extend(suggestion_value)
                    else:
                        transform_function = elasticsearch_annotation.value
                        if transform_function is not None:
                            return transform_function(section)

            return value

        def exclude(property_, section):
            if property_ not in self.indexed_properties:
                return True

            return False

        kwargs: Dict[str, Any] = dict(
            with_meta=False,
            include_defaults=True,
            include_derived=True,
            resolve_references=True,
            exclude=exclude,
            transform=transform
        )

        result = root.m_to_dict(**kwargs)

        # Add the collected suggestion values
        for path, value in suggestions.items():
            try:
                parts = path.split("/")
                section = result
                for part in parts[:-1]:
                    if part == "":
                        continue
                    try:
                        part = int(part)
                    except ValueError:
                        pass
                    section = section[part]
                section[parts[-1]] = value
            except KeyError:
                # TODO This typically happens when a suggestion is stored in a
                # referenced object that is in the metadata/results sections,
                # e.g. referenced authored that are stored in EELS DB
                # measurements
                logger = utils.get_logger(__name__, doc_type=self.name)
                logger.warn("could not add suggestion to es", path=path)

        # TODO deal with metadata
        metadata = result.get('metadata')
        if metadata is not None:
            del(result['metadata'])
            result.update(**metadata)

        return result

    def create_mapping(
            self, section_def: Section, prefix: str = None,
            auto_include_subsections: bool = False):
        '''
        Creates an Elasticsearch mapping for the given root section. It traverses all
        sub-sections to create the mapping. It will not create the mapping for nested
        documents. These have to be created manually (e.g. by :func:`create_indices`).
        Will override the existing mapping.

        Arguments:
            section_def: The section definition to create a mapping for.
            prefix: The qualified name of the section within the search index. This
                is used to create the qualified names of quantities and sub-sections.
            auto_include_subsections: Considers all sub and sub sub sections regardless
                of any annotation in the sub section definitions. By default only
                sub sections with elasticsearch annotation are traversed.
        '''
        mappings: Dict[str, Any] = {}

        if self == material_type and prefix is None:
            mappings['n_entries'] = {'type': 'integer'}

        for quantity_def in section_def.all_quantities.values():
            elasticsearch_annotations = quantity_def.m_get_annotations(Elasticsearch, as_list=True)
            for elasticsearch_annotation in elasticsearch_annotations:
                if self != entry_type and elasticsearch_annotation.doc_type != self:
                    continue

                if prefix is None:
                    qualified_name = quantity_def.name
                else:
                    qualified_name = f'{prefix}.{quantity_def.name}'
                if elasticsearch_annotation.suggestion:
                    self.suggestions[qualified_name] = elasticsearch_annotation
                is_section_reference = isinstance(quantity_def.type, Reference)
                is_section_reference &= not isinstance(quantity_def.type, QuantityReference)
                if is_section_reference:
                    # Treat referenced sections as sub-sections
                    assert quantity_def.type.target_section_def is not None
                    # TODO e.g. viewers, entry_coauthors, etc. ... should be treated as multiple inner docs
                    # assert quantity_def.is_scalar

                    reference_mapping = self.create_mapping(
                        cast(Section, quantity_def.type.target_section_def),
                        prefix=qualified_name)
                    if len(reference_mapping['properties']) > 0:
                        mappings[quantity_def.name] = reference_mapping
                else:
                    mapping = mappings.setdefault(elasticsearch_annotation.property_name, {})
                    fields = elasticsearch_annotation.fields
                    if len(fields) > 0:
                        mapping.setdefault('fields', {}).update(**fields)

                    else:
                        mapping.update(**elasticsearch_annotation.mapping)

                self.indexed_properties.add(quantity_def)
                self._register(elasticsearch_annotation, prefix)

        for sub_section_def in section_def.all_sub_sections.values():
            annotation = sub_section_def.m_get_annotations(Elasticsearch)
            if annotation is None and not auto_include_subsections:
                continue

            assert not isinstance(annotation, list), \
                'sub sections can only have one elasticsearch annotation'
            continue_with_auto_include_subsections = auto_include_subsections or (
                False if annotation is None else annotation.auto_include_subsections)

            if prefix is None:
                qualified_name = sub_section_def.name
            else:
                qualified_name = f'{prefix}.{sub_section_def.name}'

            # TODO deal with metadata
            qualified_name = re.sub(r'\.?metadata', '', qualified_name)
            qualified_name = None if qualified_name == '' else qualified_name

            sub_section_mapping = self.create_mapping(
                sub_section_def.sub_section, prefix=qualified_name,
                auto_include_subsections=continue_with_auto_include_subsections)

            nested = annotation is not None and annotation.nested
            if nested:
                sub_section_mapping['type'] = 'nested'

            if len(sub_section_mapping['properties']) > 0:
                if sub_section_def.name == 'metadata':
                    mappings.update(**sub_section_mapping['properties'])
                else:
                    mappings[sub_section_def.name] = sub_section_mapping
                self.indexed_properties.add(sub_section_def)
                if nested and qualified_name not in self.nested_object_keys:
                    self.nested_object_keys.append(qualified_name)

                    search_quantity = SearchQuantity(annotation=annotation, doc_type=self, prefix=prefix)
                    self.nested_sections.append(search_quantity)
                    self.nested_object_keys.sort(key=lambda item: len(item))

        self.mapping = dict(properties=mappings)
        return self.mapping

    def _register(self, annotation, prefix):
        search_quantity = SearchQuantity(annotation=annotation, doc_type=self, prefix=prefix)
        name = search_quantity.qualified_name

        assert name not in self.quantities or self.quantities[name] == search_quantity, \
            'Search quantity names must be unique: %s' % name

        self.quantities[name] = search_quantity

        if annotation.metrics is not None:
            for name, metric in annotation.metrics.items():
                assert name not in self.metrics, 'Metric names must be unique: %s' % name
                self.metrics[name] = (metric, search_quantity)

        if self == entry_type:
            annotation.search_quantity = search_quantity

    def __repr__(self):
        return self.name


class Index():
    '''
    Allows to access an Elasticsearch index. It forwards method calls to Python's
    Elasticsearch package for Elasticsearch document APIs like search, index, get, mget,
    bulk, etc. It adds the necessary doc_type and index parameters for you.

    Arguments:
        doc_type: The :class:`DocumentType` instance that describes the document type
            of this index.
        index_config_key: The ``nomad.config.elastic`` config key that holds the name
            for this index.

    Attributes:
        elastic_client: The used Elasticsearch Python package client.
        index_name: The name of the index in Elasticsearch.
    '''
    def __init__(self, doc_type: DocumentType, index_config_key: str):
        self.doc_type = doc_type
        self.index_config_key = index_config_key

    def __elasticsearch_operation(self, name: str, *args, **kwargs):
        if 'index' not in kwargs:
            kwargs['index'] = self.index_name

        results = getattr(self.elastic_client, name)(*args, **kwargs)
        return results

    @property
    def index_name(self):
        return getattr(config.elastic, self.index_config_key)

    @property
    def elastic_client(self):
        from nomad.infrastructure import elastic_client
        return elastic_client

    def __getattr__(self, name):
        if name not in ['get', 'index', 'mget', 'bulk', 'search']:
            return super().__getattribute__(name)

        def wrapper(*args, **kwargs):
            return self.__elasticsearch_operation(name, *args, **kwargs)

        return wrapper

    def create_index(self, upsert: bool = False):
        ''' Initially creates the index with the mapping of its document type. '''
        assert self.doc_type.mapping is not None, 'The mapping has to be created first.'
        logger = utils.get_logger(__name__, index=self.index_name)
        if not self.elastic_client.indices.exists(index=self.index_name):
            # TODO the settings emulate the path_analyzer used in the v0 elasticsearch_dsl
            # based index, configured by nomad.datamodel.datamodel::path_analyzer
            self.elastic_client.indices.create(index=self.index_name, body={
                'settings': {
                    'analysis': {
                        'analyzer': {
                            'path_analyzer': {'tokenizer': 'path_tokenizer', 'type': 'custom'}
                        },
                        'tokenizer': {
                            'path_tokenizer': {'pattern': '/', 'type': 'pattern'}
                        }
                    }
                },
                'mappings': self.doc_type.mapping
            })
            logger.info('elasticsearch index created')
        elif upsert:
            self.elastic_client.indices.put_mapping(
                index=self.index_name,
                body=self.doc_type.mapping)
            logger.info('elasticsearch index updated')
        else:
            logger.info('elasticsearch index exists')

    def delete(self):
        if self.elastic_client.indices.exists(index=self.index_name):
            self.elastic_client.indices.delete(index=self.index_name)

    def refresh(self):
        self.elastic_client.indices.refresh(index=self.index_name)


# TODO type 'doc' because it's the default used by elasticsearch_dsl and the v0 entries index.
# 'entry' would be more descriptive.
entry_type = DocumentType('doc', id_field='entry_id')
material_type = DocumentType('material', id_field='material_id')
material_entry_type = DocumentType('material_entry', id_field='entry_id')

entry_index = Index(entry_type, index_config_key='entries_index')
material_index = Index(material_type, index_config_key='materials_index')


def get_tokenizer(regex):
    '''Returns a function that tokenizes a given string using the provided
    regular epression.
    '''
    def tokenizer(value):
        tokens = []
        if value:
            tokens.append(value)
            for match in re.finditer(regex, value):
                if (match):
                    token = value[match.end():]
                    if token != '':
                        # Notice how we artificially extend the token by taking the
                        # prefix and adding it at the end. This way the token
                        # remains unique so that it will be returned by
                        # ElasticSearch when "skip_duplicates" is used in the
                        # query.
                        tokens.append(f'{token} {value[:match.end()]}')
        return tokens

    return tokenizer


tokenizer_default = get_tokenizer(r'[_\s\.\/]+')


class Elasticsearch(DefinitionAnnotation):
    '''
    A metainfo annotation for quantity definitions. This annotation can be used multiple
    times on the same quantity (e.g. to define Elasticsearch fields with differrent mapping
    types). Each annotation will create a field in the respective elasticsearch document type.

    This annotation has to be used on all sub sections that lead to quantities that should
    be included. On sub sections an inner document mapping is applied and all other
    arguments are ignored.

    Arguments:
        doc_type: An additional document type: ``material_type`` or ``material_entry_type``.
            All quantities with this annotation are automatically placed in ``entry_type``.
        mapping: The Elasticsearch mapping for the underlying elasticsearch field. The
            default depends on the quantity type. You can provide the elasticsearch type
            name, a full dictionary with additional elasticsearch mapping parameters, or
            an elasticsearch_dsl mapping object.
        field: Allows to specify sub-field name. There has to be another annotation on the
            same quantity with the default name. The custom field name is concatenated
            to the default. This will create an additional mapping for this
            quantity. In queries this can be used like an additional field, but the
            quantity is only stored once (under the quantity name) in the source document.
        value:
            A callable that is applied to the containering section to get a value for
            this quantity when saving the section in the elastic search index. By default
            this will be the serialized quantity value.
        index:
            A boolean that indicates if this quantity should be indexed or merely be
            part of the elastic document ``_source`` without being indexed for search.
        values:
            If the quantity is used in aggregations for a fixed set of values,
            use this parameter to preset these values. On aggregation, elasticsearch
            will only return values that exist in the search results. This allows to
            create 0 statistic values and return consistent set of values. If the underlying
            quantity is an Enum, the values are determined automatically.
        default_aggregation_size:
            The of values to return by default if this quantity is used in aggregation.
            If no value is given and there are not fixed value, 10 will be used.
        metrics:
            If the quantity is used as a metric for aggregating, this has to
            be used to define a valid elasticsearch metrics aggregations, e.g.
            'sum' or 'cardinality'. It is a dictionary with metric name as key,
            and elasticsearch aggregation name as values.
        many_all:
            Multiple values can be used to search based on a property. If no operator
            is given by the user, a logical or is used, i.e., all those entries that
            have at least one the values are selected (any operator). Set many_all to true
            to alter the default behavior to a logical and (all operator). If the user
            provides an operator, the provided operator is used regardless. Usually
            many_all is only sensible for properties that can have multiple value per
            entry (e.g. elements).
        auto_include_subsections:
            If true all sub and sub sub sections are considered for search even if
            there are no elasticsearch annotations in the sub section definitions.
            By default only sub sections with elasticsearch annotation are considered
            during index mapping creation.
        nested:
            If true the section is mapped to elasticsearch nested object and all queries
            become nested queries. Only applicable to sub sections.
        suggestion:
            Controls the suggestions that are built for this field. Leave
            undefined if no suggestions are required. Can be a custom callable
            that transforms a string into a list of suggestion values, or one
            of the preset strings:
            - simple: Only the value is stored as an ES field.
            - default: The value is split into tokens using whitespace, dot and
              forward slash
        variants:
            A callable that is applied to a search value to get a list of
            alternative forms of the input. Used to augment the available
            suggestions with alternative forms.
        normalizer:
            A callable that is used to transform the search input when
            targeting this field. Note that this does not affect the way the
            value is indexed.
        es_query: The Elasticsearch query type that is used when querying for the annotated
            quantity, e.g. match, term, match_phrase. Default is 'match'.

    Attributes:
        name:
            The name of the quantity (plus additional field if set).
        search_quantity: The entry type SearchQuantity associated with this annoation.
    '''
    def __init__(
            self,
            doc_type: DocumentType = entry_type,
            mapping: Union[str, Dict[str, Any]] = None,
            field: str = None,
            es_field: str = None,
            value: Callable[[MSectionBound], Any] = None,
            index: bool = True,
            values: List[str] = None,
            default_aggregation_size: int = None,
            metrics: Dict[str, str] = None,
            many_all: bool = False,
            auto_include_subsections: bool = False,
            nested: bool = False,
            suggestion: Union[str, Callable[[MSectionBound], Any]] = None,
            variants: Union[Callable[[str], List[str]]] = None,
            normalizer: Callable[[Any], Any] = None,
            es_query: str = 'match',
            _es_field: str = None):

        # TODO remove _es_field if it is not necessary anymore to enforce a specific mapping
        # for v0 compatibility
        if suggestion:
            if doc_type != entry_type:
                raise ValueError('Suggestions should only be stored in the entry index.')
            for arg in [field, mapping, es_field, _es_field]:
                if arg is not None:
                    raise ValueError(f'You cannot modify the way suggestions are mapped or named.')
            # If no tokenizer is specified, the suggestion is stored as a field
            # that holds only the original value.
            if suggestion == 'simple':
                field = 'suggestion'
            elif suggestion == 'default':
                value = tokenizer_default
            elif callable(suggestion):
                value = suggestion
            else:
                raise ValueError(
                    'Please provide the suggestion as one of the predefined '
                    'shortcuts, False or a custom callable.'
                )

        if variants and not callable(variants):
            raise ValueError('Please provide the variants as a custom callable.')
        self.variants = variants

        if normalizer and not callable(normalizer):
            raise ValueError('Please provide the normalizer as a custom callable.')
        self.normalizer = normalizer

        self._custom_mapping = mapping
        self.field = field
        self.es_query = es_query
        self._es_field = field if _es_field is None else _es_field
        self.doc_type = doc_type
        self.value = value
        self.index = index
        self._mapping: Dict[str, Any] = None

        self.default_aggregation_size = default_aggregation_size
        self.values = values
        self.metrics = metrics
        self.many_all = many_all

        self.auto_include_subsections = auto_include_subsections
        self.nested = nested
        self.suggestion = suggestion

        self.search_quantity = None

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, value):
        self._values = value
        if self.default_aggregation_size is None and self._values is not None:
            self.default_aggregation_size = len(self._values)

    @property
    def mapping(self) -> Dict[str, Any]:
        if self._mapping is not None:
            return self._mapping

        if self.suggestion:
            from elasticsearch_dsl import Completion
            # The standard analyzer will retain numbers unlike the simple
            # analyzer which is the default.
            self._mapping = Completion(analyzer='standard').to_dict()
            return self._mapping

        if self._custom_mapping is not None:
            from elasticsearch_dsl import Field

            if isinstance(self._custom_mapping, Field):
                self._mapping = self._custom_mapping.to_dict()
            elif isinstance(self._custom_mapping, str):
                self._mapping = dict(type=self._custom_mapping)
            else:
                self._mapping = self._custom_mapping

            return self._mapping

        def compute_mapping(quantity: Quantity) -> Dict[str, Any]:
            if quantity.type == str:
                return dict(type='keyword')
            elif quantity.type in [float, np.float64]:
                return dict(type='double')
            elif quantity.type == np.float32:
                return dict(type='float')
            elif quantity.type in [np.int64]:
                return dict(type='long')
            elif quantity.type in [int, np.int32]:
                return dict(type='integer')
            elif quantity.type == bool:
                return dict(type='boolean')
            elif quantity.type == Datetime:
                return dict(type='date')
            elif quantity.type == Unit:
                return dict(type='keyword')
            elif isinstance(quantity.type, QuantityReference):
                return compute_mapping(quantity.type.target_quantity_def)
            elif isinstance(quantity.type, Reference):
                raise NotImplementedError('Resolving section references is not supported.')
            elif isinstance(quantity.type, MEnum):
                return dict(type='keyword')
            else:
                raise NotImplementedError(
                    'Quantity type %s for quantity %s is not supported.' % (quantity.type, quantity))

        self._mapping = compute_mapping(cast(Quantity, self.definition))

        if not self.index:
            self._mapping['index'] = False

        return self._mapping

    @property
    def fields(self) -> Dict[str, Any]:
        if self._es_field == '' or self._es_field is None:
            return {}

        return {
            self._es_field: self.mapping
        }

    @property
    def property_name(self) -> str:
        if self.suggestion and not self.field:
            return f'{self.definition.name}__suggestion'
        return self.definition.name

    @property
    def name(self) -> str:
        if self.field is not None:
            return f'{self.property_name}.{self.field}'
        elif self.suggestion:
            return f'{self.property_name}__suggestion'
        else:
            return self.property_name

    def __repr__(self):
        if self.definition is None:
            return super().__repr__()

        return f'Elasticsearch({self.definition})'

    def m_to_dict(self):
        if self.search_quantity:
            return self.search_quantity.qualified_name
        else:
            return self.name


class SearchQuantity():
    '''
    This is used to represent search quantities. It is different from a metainfo quantity
    because the same metainfo quantity can appear multiple times at different places in
    an archive (an search index document). A search quantity is uniquely identified by
    a qualified name that pin points its place in the sub-section hierarchy.

    Arguments:
        annotation: The elasticsearch annotation that this search quantity is based on.
        doc_type: The elasticsearch document type that this search quantity appears in.
        prefix: The prefix to build the full qualified name for this search quantity.

    Attributes:
        qualified_field:
            The full qualified name of the resulting elasticsearch field in the entry
            document type. This will be the quantity name (plus additional
            field or suggestion postfix if set) with subsection names up to the
            root of the metainfo data.
        search_field:
            The full qualified name of the field in the elasticsearch index.
        qualified_name:
            Same name as qualified_field. This will be used to address the search
            property in our APIs.
        definition: The metainfo quantity definition that this search quantity is based on
        aggregateable:
            A boolean that determines, if this quantity can be used in aggregations.
    '''
    def __init__(self, annotation: Elasticsearch, doc_type: DocumentType, prefix: str):
        self.annotation = annotation
        self.doc_type = DocumentType

        qualified_field = self.annotation.definition.name

        if prefix is not None:
            qualified_field = f'{prefix}.{qualified_field}'

        if annotation.suggestion:
            qualified_field = f'{qualified_field}__suggestion'

        self.search_field = qualified_field
        if not(annotation._es_field == '' or annotation._es_field is None):
            self.search_field = f'{qualified_field}.{annotation._es_field}'

        if annotation.field is not None:
            qualified_field = f'{qualified_field}.{annotation.field}'

        self.qualified_field = qualified_field
        self.qualified_name = qualified_field

    @property
    def definition(self):
        return self.annotation.definition

    @property
    def aggregateable(self):
        if isinstance(self.definition.type, Reference):
            return False

        field_type = self.annotation.mapping['type']
        return field_type == 'keyword' or field_type == 'boolean'

    def __repr__(self):
        if self.definition is None:
            return super().__repr__()

        return f'SearchQuantity({self.qualified_field})'

    def __getattr__(self, name):
        return getattr(self.annotation, name)


def create_indices(entry_section_def: Section = None, material_section_def: Section = None):
    '''
    Creates the mapping for all document types and creates the indices in Elasticsearch.
    The indices must not exist already. Prior created mappings will be replaced.
    '''
    if entry_section_def is None:
        from nomad.datamodel import EntryArchive
        entry_section_def = EntryArchive.m_def

    if material_section_def is None:
        from nomad.datamodel.results import Material
        material_section_def = Material.m_def

    entry_type._reset()
    material_type._reset()
    material_entry_type._reset()

    entry_type.create_mapping(entry_section_def)
    material_type.create_mapping(material_section_def, auto_include_subsections=True)
    material_entry_type.create_mapping(entry_section_def, prefix='entries')

    # Here we manually add the material_entry_type mapping as a nested field
    # inside the material index. We also need to manually specify the
    # additional nested fields that come with this: the entries + all
    # nested_object_keys from material_entry_type. Notice that we need to sort
    # the list: the API expects a list sorted by name length in ascending
    # order.
    material_entry_type.mapping['type'] = 'nested'
    material_type.mapping['properties']['entries'] = material_entry_type.mapping
    material_type.nested_object_keys += ['entries'] + material_entry_type.nested_object_keys
    material_type.nested_object_keys.sort(key=lambda item: len(item))

    entry_index.create_index(upsert=True)  # TODO update the existing v0 index
    material_index.create_index()


def delete_indices():
    entry_index.delete()
    material_index.delete()


def index_entry(entry: MSection, **kwargs):
    '''
    Upserts the given entry in the entry index. Optionally updates the materials index
    as well.
    '''
    index_entries([entry], **kwargs)


def index_entries_with_materials(entries: List, refresh: bool = False):
    index_entries(entries, refresh=refresh)
    update_materials(entries, refresh=refresh)


def index_entries(entries: List, refresh: bool = False):
    '''
    Upserts the given entries in the entry index. Optionally updates the materials index
    as well.
    '''
    # split into reasonably sized problems
    if len(entries) > config.elastic.bulk_size:
        for entries_part in [entries[i:i + config.elastic.bulk_size] for i in range(0, len(entries), config.elastic.bulk_size)]:
            index_entries(entries_part, refresh=refresh)
        return

    if len(entries) == 0:
        return

    logger = utils.get_logger('nomad.search', n_entries=len(entries))

    with utils.timer(logger, 'prepare bulk index of entries actions and docs'):
        actions_and_docs = []
        for entry in entries:
            try:
                entry_index_doc = entry_type.create_index_doc(entry)

                actions_and_docs.append(dict(index=dict(_id=entry['entry_id'])))
                actions_and_docs.append(entry_index_doc)
            except Exception as e:
                logger.error('could not create entry index doc', entry_id=entry['entry_id'], exc_info=e)

        timer_kwargs: Dict[str, Any] = {}
        try:
            import json
            timer_kwargs['size'] = len(json.dumps(actions_and_docs))
            timer_kwargs['n_actions'] = len(actions_and_docs)
        except Exception:
            pass

    with utils.timer(
        logger, 'perform bulk index of entries',
        lnr_event='failed to bulk index entries',
        **timer_kwargs
    ):
        entry_index.bulk(
            body=actions_and_docs, refresh=refresh,
            timeout=f'{config.elastic.bulk_timeout}s',
            request_timeout=config.elastic.bulk_timeout)


def update_materials(entries: List, refresh: bool = False):
    # split into reasonably sized problems
    if len(entries) > config.elastic.bulk_size:
        for entries_part in [entries[i:i + config.elastic.bulk_size] for i in range(0, len(entries), config.elastic.bulk_size)]:
            update_materials(entries_part, refresh=refresh)
        return

    if len(entries) == 0:
        return

    logger = utils.get_logger('nomad.search', n_entries=len(entries))

    def get_material_id(entry):
        material_id = None
        try:
            material_id = entry.results.material.material_id
        except AttributeError:
            pass
        return material_id

    # Get all entry and material ids.
    entry_ids, material_ids = set(), set()
    entries_dict = {}
    for entry in entries:
        entries_dict[entry.entry_id] = entry
        entry_ids.add(entry.entry_id)
        material_id = get_material_id(entry)
        if material_id is not None:
            material_ids.add(material_id)

    logger = logger.bind(n_materials=len(material_ids))

    # Get existing materials for entries' material ids (i.e. the entry needs to be added
    # or updated).
    with utils.timer(logger, 'get existing materials', lnr_event='failed to get existing materials'):
        if material_ids:
            elasticsearch_results = material_index.mget(
                body={
                    'docs': [dict(_id=material_id) for material_id in material_ids]
                },
                request_timeout=config.elastic.bulk_timeout)
            existing_material_docs = [
                doc['_source'] for doc in elasticsearch_results['docs'] if '_source' in doc]
        else:
            existing_material_docs = []

    # Get old materials that still have one of the entries, but the material id has changed
    # (i.e. the materials where entries need to be removed due entries having different
    # materials now).
    with utils.timer(logger, 'get old materials', lnr_event='failed to get old materials'):
        elasticsearch_results = material_index.search(body={
            'size': len(entry_ids),
            'query': {
                'bool': {
                    'must': {
                        'nested': {
                            'path': 'entries',
                            'query': {
                                'terms': {
                                    'entries.entry_id': list(entry_ids)
                                }
                            }
                        }
                    },
                    'must_not': {
                        'terms': {
                            'material_id': list(material_ids)
                        }
                    }
                }
            }
        })
        old_material_docs = [hit['_source'] for hit in elasticsearch_results['hits']['hits']]

    # Compare and create the appropriate materials index actions
    # First, we go through the existing materials. The following cases need to be covered:
    # - an entry needs to be updated within its existing material (standard case)
    # - an entry needs to be added to an existing material (new entry case)
    # - there is an entry with no existing material (new material case)
    # - there is an entry that moves from one existing material to another (super rare
    #   case where an entry's material id changed within the set of other entries' material ids)
    # This n + m complexity with n=number of materials and m=number of entries

    # We create lists of bulk operations. Each list only contains enough materials to
    # have the ammount of entries in all these materials roughly match the desired bulk size.
    # Using materials as a measure might not be good enough, if a single material has
    # lots of nested entries.
    _actions_and_docs_bulks: List[List[Any]] = []
    _n_entries_in_bulk = [0]

    def add_action_or_doc(action_or_doc):
        if len(_actions_and_docs_bulks) == 0 or _n_entries_in_bulk[0] > config.elastic.bulk_size:
            _n_entries_in_bulk[0] = 0
            _actions_and_docs_bulks.append([])
        _actions_and_docs_bulks[-1].append(action_or_doc)
        if 'entries' in action_or_doc:
            _n_entries_in_bulk[0] = _n_entries_in_bulk[0] + len(action_or_doc['entries'])

    material_docs = []
    material_docs_dict = {}
    remaining_entry_ids = set(entry_ids)
    for material_doc in existing_material_docs:
        material_id = material_doc['material_id']
        material_docs_dict[material_id] = material_doc
        material_entries = material_doc['entries']
        material_entries_to_remove = []
        for index, material_entry in enumerate(material_entries):
            entry_id = material_entry['entry_id']
            entry = entries_dict.get(entry_id)
            if entry is None:
                # The entry was not changed.
                continue
            else:
                # Update the material, there might be slight changes even if it is made
                # from entry properties that are "material defining", e.g. changed external
                # material quantities like new AFLOW prototypes
                try:
                    material_doc.update(**material_type.create_index_doc(entry.results.material))
                except Exception as e:
                    logger.error('could not create material index doc', exc_info=e)

            new_material_id = get_material_id(entry)
            if new_material_id != material_id:
                # Remove the entry, it moved to another material. But the material cannot
                # run empty, because another entry had this material id.
                material_entries_to_remove.append(index)
            else:
                # Update the entry.
                try:
                    material_entries[index] = material_entry_type.create_index_doc(entry)
                except Exception as e:
                    logger.error('could not create material index doc', exc_info=e)
                remaining_entry_ids.remove(entry_id)
        for index in reversed(material_entries_to_remove):
            del(material_entries[index])

        add_action_or_doc(dict(index=dict(_id=material_id)))
        add_action_or_doc(material_doc)
        material_docs.append(material_doc)

    for entry_id in remaining_entry_ids:
        entry = entries_dict.get(entry_id)
        material_id = get_material_id(entry)
        if material_id is not None:
            material_doc = material_docs_dict.get(material_id)
            if material_doc is None:
                # The material does not yet exist. Create it.
                try:
                    material_doc = material_type.create_index_doc(entry.results.material)
                except Exception as e:
                    logger.error('could not create material index doc', exc_info=e)
                material_docs_dict[material_id] = material_doc
                add_action_or_doc(dict(create=dict(_id=material_id)))
                add_action_or_doc(material_doc)
                material_docs.append(material_doc)
            # The material does exist (now), but the entry is new.
            try:
                material_doc.setdefault('entries', []).append(material_entry_type.create_index_doc(entry))
            except Exception as e:
                logger.error('could not create material entry index doc', exc_info=e)

    # Second, we go through the old materials. The following cases need to be covered:
    # - the old materials are empty (standard case)
    # - an entry needs to be removed but the material still has entries (new material id case 1)
    # - an entry needs to be removed and the material is now "empty" (new material id case 2)
    for material_doc in old_material_docs:
        material_id = material_doc['material_id']
        material_entries = material_doc['entries']
        material_entries_to_remove = []
        for index, material_entry in enumerate(material_entries):
            entry_id = material_entry['entry_id']
            if entry_id in entry_ids:
                # The entry does not belong to this material anymore and needs to be removed.
                material_entries_to_remove.append(index)
        for index in reversed(material_entries_to_remove):
            del(material_entries[index])
        if len(material_entries) == 0:
            # The material is empty now and needs to be removed.
            add_action_or_doc(dict(delete=dict(_id=material_id)))
        else:
            # The material needs to be updated
            add_action_or_doc(dict(index=dict(_id=material_id)))
            add_action_or_doc(material_doc)
            material_docs.append(material_doc)

    # Third, we potentially cap the number of entries in a material. We ensure that only
    # a certain amounts of entries are stored with all metadata. The rest will only
    # have their entry id.
    all_n_entries_capped = 0
    all_n_entries = 0
    for material_doc in material_docs:
        material_entries = material_doc.get('entries', [])
        material_doc['n_entries'] = len(material_entries)
        if len(material_entries) > config.elastic.entries_per_material_cap:
            material_doc['entries'] = material_entries[0:config.elastic.entries_per_material_cap]

        all_n_entries_capped += len(material_entries)
        all_n_entries += material_doc['n_entries']

    # Execute the created actions in bulk.
    timer_kwargs: Dict[str, Any] = {}
    try:
        import json
        timer_kwargs['size'] = len(json.dumps(_actions_and_docs_bulks))
        timer_kwargs['n_actions'] = sum([len(bulk) for bulk in _actions_and_docs_bulks])
        timer_kwargs['n_entries'] = all_n_entries
        timer_kwargs['n_entries_capped'] = all_n_entries_capped
    except Exception:
        pass

    with utils.timer(
        logger, 'perform bulk index of materials',
        lnr_event='failed to bulk index materials',
        **timer_kwargs
    ):
        for bulk in _actions_and_docs_bulks:
            material_index.bulk(
                body=bulk, refresh=False,
                timeout=f'{config.elastic.bulk_timeout}s',
                request_timeout=config.elastic.bulk_timeout)

    if refresh:
        entry_index.refresh()
        material_index.refresh()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any
from fastapi import APIRouter, Depends, Path, status, HTTPException, Request
from fastapi.exception_handlers import RequestValidationError
from pydantic import BaseModel, Field

from nomad import utils
from nomad.utils import strip
from nomad.search import AuthenticationRequiredError, SearchError
from nomad.search import search, QueryValidationError
from nomad.metainfo.elasticsearch_extension import material_type, material_index

from .auth import create_user_dependency
from ..utils import create_responses
from ..models import (
    User, Owner, WithQuery,
    MetadataResponse, Metadata, MetadataPagination, MetadataRequired,
    metadata_pagination_parameters, metadata_required_parameters, QueryParameters,
    HTTPExceptionModel)


router = APIRouter()

logger = utils.get_logger(__name__)

query_parameters = QueryParameters(doc_type=material_type)

_bad_owner_response = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. The given owner requires authorization,
        but no or bad authentication credentials are given.''')}

_bad_id_response = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Material not found. The given id does not match any material.''')}


class MaterialMetadataResponse(BaseModel):
    material_id: str = Field(None)
    required: MetadataRequired = Field(None)
    data: Any = Field(
        None, description=strip('''The material metadata as dictionary.'''))


def perform_search(*args, **kwargs) -> MetadataResponse:
    kwargs.update(index=material_index)
    try:
        search_response = search(*args, **kwargs)
        search_response.es_query = None
        return search_response
    except QueryValidationError as e:
        raise RequestValidationError(errors=e.errors)
    except AuthenticationRequiredError as e:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=str(e))
    except SearchError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Elasticsearch could not process your query: %s' % str(e))


@router.post(
    '/query', tags=['materials'],
    summary='Search materials and retrieve their metadata',
    response_model=MetadataResponse,
    responses=create_responses(_bad_owner_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_entries_metadata_query(
        request: Request,
        data: Metadata,
        user: User = Depends(create_user_dependency())):

    '''
    Executes a *query* and returns a *page* of the results with *required* result data
    as well as *statistics* and *aggregated* data.

    This is the basic search operation to retrieve metadata for entries that match
    certain search criteria (`query` and `owner`). All parameters (including `query`, `owner`)
    are optional. Look at the body schema or parameter documentation for more details.

    By default the *empty* search (that returns everything) is performed. Only a small
    page of the search results are returned at a time; use `pagination` in subsequent
    requests to retrive more data. Each entry has a lot of different *metadata*, use
    `required` to limit the data that is returned.

    The `statistics` and `aggregations` keys will further allow to return statistics
    and aggregated data over all search results.
    '''

    return perform_search(
        owner=data.owner,
        query=data.query,
        pagination=data.pagination,
        required=data.required,
        aggregations=data.aggregations,
        user_id=user.user_id if user is not None else None)


@router.get(
    '', tags=['materials'],
    summary='Search materials and retrieve their metadata',
    response_model=MetadataResponse,
    responses=create_responses(_bad_owner_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_entries_metadata(
        request: Request,
        with_query: WithQuery = Depends(query_parameters),
        pagination: MetadataPagination = Depends(metadata_pagination_parameters),
        required: MetadataRequired = Depends(metadata_required_parameters),
        user: User = Depends(create_user_dependency())):
    '''
    Executes a *query* and returns a *page* of the results with *required* result data.
    This is a version of `/entries/query`. Queries work a little different, because
    we cannot put complex queries into URL parameters.

    In addition to the `q` parameter (see parameter documentation for details), you can use all NOMAD
    search quantities as parameters, e.g. `?atoms=H&atoms=O`. Those quantities can be
    used with additional operators attached to their names, e.g. `?n_atoms__gte=3` for
    all entries with more than 3 atoms. Operators are `all`, `any`, `none`, `gte`,
    `gt`, `lt`, `lte`.
    '''

    res = perform_search(
        owner=with_query.owner, query=with_query.query,
        pagination=pagination, required=required,
        user_id=user.user_id if user is not None else None)
    res.pagination.populate_urls(request)
    return res


@router.get(
    '/{material_id}', tags=['materials'],
    summary='Get the metadata of a material by its id',
    response_model=MaterialMetadataResponse,
    responses=create_responses(_bad_id_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_material_metadata(
        material_id: str = Path(..., description='The unique material id of the material to retrieve metadata from.'),
        required: MetadataRequired = Depends(metadata_required_parameters),
        user: User = Depends(create_user_dependency())):
    '''
    Retrives the material metadata for the given id.
    '''

    query = {'material_id': material_id}
    response = perform_search(
        owner=Owner.all_, query=query, required=required,
        user_id=user.user_id if user is not None else None)

    if response.pagination.total == 0:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='The entry with the given id does not exist or is not visible to you.')

    return {
        'material_id': material_id,
        'required': required,
        'data': response.data[0]
    }



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import bz2
import datetime
import gzip
import lzma
import numpy as np
import logging

from nomad.units import ureg
from nomad.datamodel.metainfo.simulation.run import Run, Program, TimeRun
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.simulation.method import Method, BasisSet
from nomad.datamodel.metainfo.simulation.calculation import Calculation, Forces, ForcesEntry
from nomad.parsing.file_parser import Quantity, TextParser

import xml.etree.ElementTree as ElementTree


def str_to_timestamp(s: str):
    return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%SZ").timestamp()


class QBallParser:

    mainfile_parser = TextParser(
        quantities=[
            Quantity(
                "atoms",
                r"symbol_ = (\w+)",
                repeats=True,
            ),
            Quantity(
                "start_time",
                r"<start_time>\s*(.*?)\s*</start_time>",
                repeats=False,
            ),
            Quantity(
                "end_time",
                r"<end_time>\s*(.*?)\s*</end_time>",
                repeats=False,
            ),
        ]
    )

    def __init__(self):
        pass

    def parse(self, mainfile, archive, logger=None):
        logger = logger if logger is not None else logging.getLogger('__name__')

        if mainfile.endswith(".gz"):
            open_file = gzip.open
        elif mainfile.endswith(".bz2"):
            open_file = bz2.open
        elif mainfile.endswith(".xz"):
            open_file = lzma.open

        with open_file(mainfile, "rt") as file:
            contents = file.read()

        self.mainfile_parser.mainfile = mainfile
        self.mainfile_parser.parse()

        run = archive.m_create(Run)
        run.program = Program(name='qball')
        run.time_run = TimeRun(
            date_start=str_to_timestamp(self.mainfile_parser.get("start_time")),
            date_end=str_to_timestamp(self.mainfile_parser.get("end_time")))

        # method
        method = run.m_create(Method)
        # TODO add dft functionals
        if "plane waves" not in contents:
            logger.error("Qball Error: Not a plane wave dft")
        else:
            basis_set = method.m_create(BasisSet)
            basis_set.type = "plane waves"

        element_tree = ElementTree.fromstring(contents)

        # system
        system = run.m_create(System)
        system.atoms = Atoms(
            labels=[atom.attrib["name"] for atom in element_tree.find("run").find("iteration").find("atomset").iter("atom")],
            positions=np.array([
                [float(pos) for pos in atom.find("position").text.split()]
                for atom in element_tree.find("run").find("iteration").find("atomset").iter("atom")]) * ureg.bohr)

        # calculation
        calculation = run.m_create(Calculation)
        calculation.forces = Forces(total=ForcesEntry(value=np.array([
            [float(force) for force in atom.find("force").text.split()]
            for atom in element_tree.find("run").find("iteration").find("atomset").iter("atom")]) * ureg.hartree / ureg.bohr))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .abinit.parser import AbinitParser
from .atk.parser import ATKParser
from .band.parser import BandParser
from .bigdft.parser import BigDFTParser
from .castep.parser import CastepParser
from .charmm.parser import CharmmParser
from .cp2k.parser import CP2KParser
from .cpmd.parser import CPMDParser
from .crystal.parser import CrystalParser
from .dmol3.parser import Dmol3Parser
from .elk.parser import ElkParser
from .exciting.parser import ExcitingParser
from .fhiaims.parser import FHIAimsParser
from .fleur.parser import FleurParser
from .fplo.parser import FploParser
from .gamess.parser import GamessParser
from .gaussian.parser import GaussianParser
from .gpaw.parser import GPAWParser
from .molcas.parser import MolcasParser
from .nwchem.parser import NWChemParser
from .octopus.parser import OctopusParser
from .onetep.parser import OnetepParser
from .orca.parser import OrcaParser
from .psi4.parser import Psi4Parser
from .qball.parser import QBallParser
from .qbox.parser import QboxParser
from .quantumespresso.parser import QuantumEspressoParser
from .siesta.parser import SiestaParser
from .turbomole.parser import TurbomoleParser
from .vasp.parser import VASPParser
from .wien2k.parser import Wien2kParser
from .yambo.parser import YamboParser



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from . import Normalizer


class MetainfoNormalizer(Normalizer):
    domain = None

    def normalize(self, logger=None) -> None:
        if logger is not None:
            self.logger = logger.bind(normalizer=self.__class__.__name__)

        if not self.entry_archive.data:
            return

        for section, _, _ in list(self.entry_archive.data.m_traverse()):
            normalize = None
            try:
                normalize = getattr(section, 'normalize')
            except Exception as e:
                pass

            if normalize:
                try:
                    normalize(self.entry_archive, logger)
                except Exception as e:
                    logger.error(
                        'could not normalize section',
                        section=section.m_def.name,
                        exc_info=e)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Commands for updating the DOS similarity of materials from an external file.
'''

# TODO If we want to keep supporting DOS similarities in the next version of
# Encyclopedia, this script needs to be updated to work with the new materials
# index.

# import os
# import json
# from nomad import infrastructure
# from nomad.datamodel.material import Material, Similarity, DOSSimilarity
# from nomad.archive import ArchiveReader, ArchiveWriter
# from typing import List
#
#
# def ingest(input_path: str, batch_size: int, verbose: bool):
#     """Used to ingest the given DOS similarity values into MongoDB.
#
#     Args:
#         input_path: Path of the msgpack file to ingest.
#         batch_size: Batch size for MongoDB bulk ingest.
#         verbose: Enable verbose output.
#     """
#     # Initialize mongo connection
#     infrastructure.setup_mongo()
#
#     bulk = Material.m_def.a_mongo.mongo_cls()._get_collection().initialize_ordered_bulk_op()  # pylint: disable=not-callable
#
#     with ArchiveReader(input_path) as reader:
#         i = 0
#         for material_id in reader:
#             material_object = reader[material_id]
#             material_dict = material_object.to_dict()
#
#             # Each entry is cycled through the material metainfo definition and
#             # the mongo annotation to validate it and only push the annotated
#             # data.
#             mongo_instance = Material.m_def.a_mongo.mongo_cls(**material_dict, _created=False)  # pylint: disable=not-callable
#             material_dict = mongo_instance.to_mongo().to_dict()
#
#             # Add an upsert to the bulk operations
#             bulk.find({'_id': material_dict["_id"]}).upsert().update_one({'$set': material_dict})
#             i += 1
#             if i % batch_size == 0:
#                 bulk.execute()
#                 bulk = Material.m_def.a_mongo.mongo_cls()._get_collection().initialize_ordered_bulk_op()  # pylint: disable=not-callable
#                 if verbose:
#                     print("{} inserted".format(i))
#
#     # Final push for the remainder
#     bulk.execute()
#     if verbose:
#         print("{} inserted".format(i))
#
#
# def update(input_dir: str, output_path: str, verbose: bool):
#     """Used to create a compact msgpack file that follows the metainfo schema
#     for materials and contains the DOS similarity values.
#
#     Args:
#         input_dir: Path of the directory containing the raw similarity filesa.
#         output_path: Path of the output msgpack file.
#         verbose: Enable verbose output.
#     """
#     # Find all valid data files in the given directory
#     similarity_files: List[str] = []
#     for filename in os.listdir(input_dir):
#         if filename.endswith(".dat"):
#             similarity_files.append(os.path.join(input_dir, filename))
#     n_files: int = len(similarity_files)
#     if n_files == 0:
#         raise ValueError("Could not find similarity files in directory: {}".format(input_dir))
#     if verbose:
#         print("{} files found".format(n_files))
#
#     # Gather the number of entries to prepare the msgpack file.
#     n_entries = 0
#     for filepath in similarity_files:
#         num_lines = sum(1 for line in open(filepath, "r"))
#         n_entries += num_lines
#     if verbose:
#         print("{} entries found".format(n_entries))
#
#     # Read the data file containing similarities
#     # i = 0
#     if verbose:
#         print("Writing msgpack file...")
#     with ArchiveWriter(output_path, n_entries, entry_toc_depth=1) as writer:
#         for filepath in similarity_files:
#             with open(filepath, "r") as f:
#                 for line in f:
#                     ientry = json.loads(line)
#
#                     for key, value in ientry.items():
#                         _, _, imaterial = key.split(":")
#
#                         # Create data according to a metainfo model
#                         material = Material()
#                         material.material_id = imaterial
#                         similarity = material.m_create(Similarity)
#                         dos_similarity = similarity.m_create(DOSSimilarity)
#                         ids = []
#                         values = []
#
#                         for jkey, similarity_value in value.items():
#                             _, _, jmaterial = jkey.split(":")
#                             ids.append(jmaterial)
#                             values.append(similarity_value)
#
#                         dos_similarity.material_ids = ids
#                         dos_similarity.values = values
#
#                         # Save as msgpack
#                         writer.add(imaterial, material.m_to_dict())
#
#     if verbose:
#         print("Finished")



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from nptyping import NDArray

from nomad import config
from nomad_dos_fingerprints import DOSFingerprint
from nomad.datamodel.metainfo.simulation.calculation import (
    Dos, DosFingerprint, BandGap)
from nomad.atomutils import get_volume

from .normalizer import Normalizer


class DosNormalizer(Normalizer):
    """Normalizer with the following responsibilities:

      - Determines highest occupied and lowest unoccupied energies for both
        spin channels.
      - Adds normalization factor for intensive (=not tied to the imulation
        cell size) DOS values which are
    """
    def normalize(self, logger=None) -> None:

        if logger is not None:
            self.logger = logger.bind(normalizer=self.__class__.__name__)

        # Do nothing if section_run is not present
        if self.section_run is None:
            return

        calculations = self.section_run.calculation
        if calculations is None:
            return

        for calc in calculations:
            # Normalize electronic DOS
            dos_electronic = calc.dos_electronic
            if dos_electronic is not None:
                energy_fermi = calc.energy.fermi if calc.energy is not None else None
                energy_highest = calc.energy.highest_occupied if calc.energy is not None else None
                energy_lowest = calc.energy.lowest_unoccupied if calc.energy is not None else None
                for dos in dos_electronic:

                    # Add normalization factor
                    number_of_atoms = self.add_normalization_factor(calc, dos)
                    if number_of_atoms is None:
                        continue

                    # Add energy references
                    dos_values_normalized = [
                        dos_total.value.magnitude * dos_total.normalization_factor.magnitude for dos_total in dos.total]
                    self.add_energy_references(dos, energy_fermi, energy_highest, energy_lowest, dos_values_normalized)

                    # Calculate the DOS fingerprint for successfully normalized DOS
                    normalization_reference = None
                    for info in dos.band_gap:
                        energy_highest = info.energy_highest_occupied
                        if energy_highest is not None:
                            if normalization_reference is None:
                                normalization_reference = energy_highest
                            else:
                                normalization_reference = max(normalization_reference, energy_highest)
                    if normalization_reference is not None:
                        dos_energies_normalized = dos.energies - normalization_reference

                        try:
                            dos_fingerprint = DOSFingerprint().calculate(
                                dos_energies_normalized.magnitude,
                                dos_values_normalized,
                                n_atoms=number_of_atoms
                            )
                        except Exception as e:
                            self.logger.error('could not generate dos fingerprint', exc_info=e)
                        else:
                            sec_dos_fingerprint = dos.m_create(DosFingerprint)
                            sec_dos_fingerprint.bins = dos_fingerprint.bins
                            sec_dos_fingerprint.indices = dos_fingerprint.indices
                            sec_dos_fingerprint.stepsize = dos_fingerprint.stepsize
                            sec_dos_fingerprint.grid_id = dos_fingerprint.grid_id
                            sec_dos_fingerprint.filling_factor = dos_fingerprint.filling_factor

            # Normalize phonon DOS
            dos_phonon = calc.dos_phonon
            if dos_phonon is not None:
                for dos in dos_phonon:
                    self.add_normalization_factor(calc, dos)

    def add_normalization_factor(self, calc, dos):
        # Perform normalization only for total dos
        if dos.total is None:
            return

        """Returns the factor that normalizes DOS values to be 1/J/atom/m^3."""
        system = calc.system_ref
        if not system or system.atoms is None:
            self.logger.error('referenced system for dos calculation could not be found')
            return
        atom_positions = system.atoms.positions
        lattice_vectors = system.atoms.lattice_vectors
        if atom_positions is None:
            self.logger.error('required quantity atom_positions is not available')
            return
        if lattice_vectors is None:
            self.logger.error('required quantity lattice_vectors is not available')
            return
        number_of_atoms = np.shape(atom_positions)[0]
        unit_cell_volume = get_volume(lattice_vectors.magnitude)
        normalization_factor = 1 / (number_of_atoms * unit_cell_volume)

        if normalization_factor:
            for dos_total in dos.total:
                dos_total.normalization_factor = normalization_factor

        return number_of_atoms

    def add_energy_references(
            self,
            dos: Dos,
            energy_fermi: float,
            energy_highest: float,
            energy_lowest: float,
            dos_values_normalized: NDArray) -> None:
        """Given the band structure and information about energy references,
        determines the energy references separately for all spin channels.
        """
        dos.energy_fermi = energy_fermi
        eref = energy_highest if energy_fermi is None else energy_fermi
        # No reference available
        if eref is None:
            self.logger.info("could not resolve energy references for dos")
            return

        # Create channel information for each spin channel and populate with
        # initial values.
        dos_total = dos.total
        n_channels = len(dos_total)
        for i_channel in range(n_channels):
            info = dos.band_gap[i_channel] if len(dos.band_gap) > i_channel else dos.m_create(BandGap)
            info.index = i_channel
            if energy_highest is not None:
                info.energy_highest_occupied = energy_highest
            if energy_lowest is not None:
                info.energy_lowest_unoccupied = energy_lowest

        # Use a reference energy (fermi or highest occupied) to determine the
        # energy references from the DOS (discretization will affect the exact
        # location).
        energy_threshold = config.normalize.band_structure_energy_tolerance
        value_threshold = 1e-8  # The DOS value that is considered to be zero
        dos_energies = dos.energies

        for i_channel in range(n_channels):
            dos_values = dos_values_normalized[i_channel]
            info = dos.band_gap[i_channel]
            fermi_idx = (np.abs(dos.energies - eref)).argmin()

            # First check that the closest dos energy to energy reference
            # is not too far away. If it is very far away, the
            # normalization may be very inaccurate and we do not report it.
            fermi_energy_closest = dos_energies[fermi_idx]
            distance = np.abs(fermi_energy_closest - eref)
            if distance.magnitude <= energy_threshold:

                # See if there are zero values close below the energy reference.
                idx = fermi_idx
                idx_descend = fermi_idx
                while True:
                    try:
                        value = dos_values[idx]
                        energy_distance = np.abs(eref - dos_energies[idx])
                    except IndexError:
                        break
                    if energy_distance.magnitude > energy_threshold:
                        break
                    if value <= value_threshold:
                        idx_descend = idx
                        break
                    idx -= 1

                # See if there are zero values close above the fermi energy.
                idx = fermi_idx
                idx_ascend = fermi_idx
                while True:
                    try:
                        value = dos_values[idx]
                        energy_distance = np.abs(eref - dos_energies[idx])
                    except IndexError:
                        break
                    if energy_distance.magnitude > energy_threshold:
                        break
                    if value <= value_threshold:
                        idx_ascend = idx
                        break
                    idx += 1

                # If there is a single peak at fermi energy, no
                # search needs to be performed.
                if idx_ascend != fermi_idx and idx_descend != fermi_idx:
                    info.energy_highest_occupied = fermi_energy_closest
                    info.energy_lowest_unoccupied = fermi_energy_closest
                    continue

                # Look for highest occupied energy below the descend index
                idx = idx_descend
                while True:
                    try:
                        value = dos_values[idx]
                    except IndexError:
                        break
                    if value > value_threshold:
                        idx = idx if idx == idx_descend else idx + 1
                        info.energy_highest_occupied = dos_energies[idx]
                        break
                    idx -= 1

                # Look for lowest unoccupied energy above idx_ascend
                idx = idx_ascend
                while True:
                    try:
                        value = dos_values[idx]
                    except IndexError:
                        break
                    if value > value_threshold:
                        idx = idx if idx == idx_ascend else idx - 1
                        info.energy_lowest_unoccupied = dos_energies[idx]
                        break
                    idx += 1



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.metainfo import MCategory, Category


class FastAccess(MCategory):
    '''
    Used to mark archive objects that need to be stored in a fast 2nd-tier storage medium,
    because they are frequently accessed via archive API.

    If applied to a sub_section, the section will be added to the fast storage. Currently
    this only works for *root* sections that are sub_sections of `EntryArchive`.

    If applied to a reference types quantity, the referenced section will also be added to
    the fast storage, regardless if the referenced section has the category or not.
    '''

    m_def = Category()
